#include <sapi.h>
#include <msg.h>
#include <tp.h>
#include <ctrans.h>
#include "ct.h"


#define CTRANS_HASH_SIZE (32768)


#define T1   5 // 500 msecs
#define T2  40 //   4 secs
#define T4  50 //   5 secs
#define T5 320 //  32 secs


static hash_t *__ht = NULL;
static mutex_t __mutex = SAPI_MUTEX_INITIALIZER;


static inline u8_t *ctrans_hash_key(const msg_t *msg, bool_t cancel)
{
	via_t *via = msg_get_topvia(msg);
	u8_t *hkey = NULL, *branch;

	if(!via) {
		CTRANS_DEBUG_PRINTF("ctrans_hash_key(): top via missing");
		return(NULL);
	}

	branch = msg_via_getparam(via, "branch");

	if(!branch) {
		CTRANS_DEBUG_PRINTF("ctrans_hash_key(): branch parameter is missing");
		return(NULL);
	}

	if(msg->request) {
		if(cancel)
			hkey = sapi_sprintf("%s|%s", branch, "INVITE");
		else
			hkey = sapi_sprintf("%s|%s", branch, msg->request->method);
	}
	else if(msg->response)
		hkey = sapi_sprintf("%s|%s", branch, msg->cseq->method);

	if(!hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_hash_key(): hkey is NULL");
	}

	return(hkey);
}


static void ctrans_tp_err(const u8_t *reason, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_tp_err(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		CTRANS_DEBUG_PRINTF("ctrans_tp_err(): ctrans not found");
		return;
	}

	sapi_mutex_lock(&__mutex);
	sapi_hash_del(__ht, hkey);
	sapi_mutex_unlock(&__mutex);

	if(ct->errcb)
		ct->errcb(ct->req, reason, ct->arg, ct->sarg);

	ct_deref(ct);
}


static inline void ctrans_tx(ct_t *ct, const msg_t *req)
{
	if(!ct) {
		CTRANS_DEBUG_PRINTF("ctrans_tx(): c is NULL");
		return;
	}

	tp_send(ct->tp, req, ctrans_tp_err, ct->hkey);
	ct_inc_ntx(ct);
}


static void ctrans_retrans_i(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_retrans_i(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		// CTRANS_DEBUG_PRINTF("ctrans_retrans_i(): ctrans not found");
		return;
	}

	if(ct_get_state(ct) == CALLING) {
		ctrans_tx(ct, ct->req);
		sapi_oneshot(T1<<ct_get_ntx(ct), ctrans_retrans_i, NULL, ct->hkey);
	}

	ct_deref(ct);
}


static void ctrans_timeout_i(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_timeout_i(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		// CTRANS_DEBUG_PRINTF("ctrans_timeout_i(): ctrans not found");
		return;
	}

	if(ct_get_state(ct) == CALLING) {

		sapi_mutex_lock(&__mutex);
		sapi_hash_del(__ht, hkey);
		sapi_mutex_unlock(&__mutex);

		if(ct->errcb)
			ct->errcb(ct->req, "INVITE Request Timeout", ct->arg, ct->sarg);
	}

	ct_deref(ct);
}


static void ctrans_timeout_cancel_i(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_timeout_cancel_i(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		// CTRANS_DEBUG_PRINTF("ctrans_timeout_cancel_i(): ctrans not found");
		return;
	}

	if(ct_get_state(ct) == PROCEEDING) {

		sapi_mutex_lock(&__mutex);
		sapi_hash_del(__ht, hkey);
		sapi_mutex_unlock(&__mutex);

		if(ct->errcb)
			ct->errcb(ct->req, "Canceled INVITE Request Timeout", ct->arg, ct->sarg);
	}

	ct_deref(ct);
}


static void ctrans_zombie_i(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_zombie_i(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		// CTRANS_DEBUG_PRINTF("ctrans_zombie_i(): ctrans not found");
		return;
	}

	if(ct_get_state(ct) == COMPLETED) {

		sapi_mutex_lock(&__mutex);
		sapi_hash_del(__ht, hkey);
		sapi_mutex_unlock(&__mutex);
	}


	ct_deref(ct);
}


static void ctrans_retrans_ni(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_retrans_ni(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		//  CTRANS_DEBUG_PRINTF("ctrans_retrans_ni(): ctrans not found");
		return;
	}

	if(ct_get_state(ct) == TRYING) {

		ctrans_tx(ct, ct->req);

		sapi_oneshot(sapi_min(T1<<ct_get_ntx(ct), T2), ctrans_retrans_ni, NULL, ct->hkey);
	}
	else if(ct_get_state(ct) == PROCEEDING) {

		ctrans_tx(ct, ct->req);

		sapi_oneshot(T2, ctrans_retrans_ni, NULL, ct->hkey);
	}

	ct_deref(ct);
}


static void ctrans_timeout_ni(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_timeout_ni(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		// CTRANS_DEBUG_PRINTF("ctrans_timeout_ni(): ctrans not found");
		return;
	}

	if((ct_get_state(ct) == TRYING || ct_get_state(ct) == PROCEEDING)) {

		sapi_mutex_lock(&__mutex);
		sapi_hash_del(__ht, hkey);
		sapi_mutex_unlock(&__mutex);

		if(ct->errcb)
			ct->errcb(ct->req, "NON-INVITE Request Timeout", ct->arg, ct->sarg);
	}

	ct_deref(ct);
}


static void ctrans_zombie_ni(void *arg, u8_t *hkey)
{
	ct_t *ct;

	if(!__ht || !hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_zombie_ni(): __ht or hkey is NULL");
		return;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		// CTRANS_DEBUG_PRINTF("ctrans_zombie_ni(): ctrans not found");
		return;
	}

	if(ct_get_state(ct) == COMPLETED) {

		sapi_mutex_lock(&__mutex);
		sapi_hash_del(__ht, hkey);
		sapi_mutex_unlock(&__mutex);
	}

	ct_deref(ct);
}


static inline void ctrans_resp_i(ct_t *ct, const msg_t *resp)
{
	/* 1XX */
	if(resp->response->statuscode[0] == '1') {

		if(ct_get_state(ct) == CALLING) {

			ct_set_state(ct, PROCEEDING);

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);

			CTRANS_DEBUG_PRINTF("ctrans_resp_i(): CALLING -> PROCEEDING");
		}
		else if(ct_get_state(ct) == PROCEEDING) {

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);
		}
	}
	/* 2XX */
	else if(resp->response->statuscode[0] == '2') {

		if(ct_get_state(ct) == CALLING || ct_get_state(ct) == PROCEEDING) {

			sapi_mutex_lock(&__mutex);
			sapi_hash_del(__ht, ct->hkey);
			sapi_mutex_unlock(&__mutex);

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);

			if(ct_get_state(ct) == CALLING)
				CTRANS_DEBUG_PRINTF("ctrans_resp_i(): CALLING -> TERMINATED");
			else
				CTRANS_DEBUG_PRINTF("ctrans_resp_i(): PROCEEDING -> TERMINATED");
		}
	}
	/* 300-699 */
	else {
		if(ct_get_state(ct) == CALLING || ct_get_state(ct) == PROCEEDING) {

			if(ct_get_state(ct) == CALLING)
				CTRANS_DEBUG_PRINTF("ctrans_resp_i(): CALLING -> COMPLETED");
			else
				CTRANS_DEBUG_PRINTF("ctrans_resp_i(): PROCEEDING -> COMPLETED");

			ct_set_state(ct, COMPLETED);

			/* Construct ACK and send */
			ct->ack = msg_ack(ct->req, resp);
			ctrans_tx(ct, ct->ack);

			if(ct->tp->proto == UDP) {
				sapi_oneshot(T5, ctrans_zombie_i, NULL, ct->hkey);
			}
			else {
				sapi_mutex_lock(&__mutex);
				sapi_hash_del(__ht, ct->hkey);
				sapi_mutex_unlock(&__mutex);

				CTRANS_DEBUG_PRINTF("ctrans_resp_i(): COMPLETED -> TERMINATED");
			}

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);

		}
		else if(ct_get_state(ct) == COMPLETED) {

			/* Send ACK */
			ctrans_tx(ct, ct->ack);
			CTRANS_DEBUG_PRINTF("ctrans_resp_i(): Final response retransmitted");
		}
	}
}


static inline void ctrans_resp_ni(ct_t *ct, const msg_t *resp)
{
	/* 1XX */
	if(resp->response->statuscode[0] == '1') {

		if(ct_get_state(ct) == TRYING) {

			ct_set_state(ct, PROCEEDING);

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);

			CTRANS_DEBUG_PRINTF("ctrans_resp_ni(): TRYING -> PROCEEDING");
		}
		else if(ct_get_state(ct) == PROCEEDING) {

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);
		}
	}
	/* 200-699 */
	else {
		if(ct_get_state(ct) == TRYING || ct_get_state(ct) == PROCEEDING) {

			if(ct_get_state(ct) == TRYING)
				CTRANS_DEBUG_PRINTF("ctrans_resp_ni(): TRYING -> COMPLETED");
			else
				CTRANS_DEBUG_PRINTF("ctrans_resp_ni(): PROCEEDING -> COMPETED");

			ct_set_state(ct, COMPLETED);

			if(ct->tp->proto == UDP) {
				sapi_oneshot(T4, ctrans_zombie_ni, NULL, ct->hkey);
			}
			else {
				sapi_mutex_lock(&__mutex);
				sapi_hash_del(__ht, ct->hkey);
				sapi_mutex_unlock(&__mutex);

				CTRANS_DEBUG_PRINTF("ctrans_resp_ni(): COMPLETED -> TERMINATED");
			}

			if(ct->respcb)
				ct->respcb(ct->req, resp, ct->arg, ct->sarg);
		}
	}
}


void ctrans_init(void)
{
	if(__ht)
		return;

	__ht = sapi_hash_alloc(CTRANS_HASH_SIZE);
	sapi_logf("Client transaction service is up and running");
}


void ctrans_req(const tp_t *tp, const msg_t *req, ctrans_resp_t respcb, ctrans_err_t errcb, void *arg, const u8_t *sarg)
{
	u8_t *hkey;
	ct_t *ct;
	timeout_t retrans, timeout;
	ret_t ret;

	if(!__ht || !tp || !req || !req->request) {
		CTRANS_DEBUG_PRINTF("ctrans_req(): __ht, tp, req or req->request is NULL");
		return;
	}

	hkey = ctrans_hash_key(req, FALSE);
	if(!hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_req(): hkey is NULL");
		return;
	}

	ct = ct_alloc();

	ct->hkey       = hkey;
	ct->tp         = tp_ref(tp);
	ct->req        = msg_ref(req);
	ct->respcb     = respcb;
	ct->errcb      = errcb;
	ct->arg        = arg;
	ct->sarg       = sapi_strdup(sarg);

	if(sapi_strcmp(req->request->method, "INVITE")) {
		ct->state = CALLING;
		retrans  = ctrans_retrans_i;
		timeout  = ctrans_timeout_i;
	}
	else {
		ct->state = TRYING;
		retrans  = ctrans_retrans_ni;
		timeout  = ctrans_timeout_ni;
	}

	sapi_mutex_lock(&__mutex);
	ret = sapi_hash_ins(__ht, ct->hkey, ct_ref(ct), ct_deref);
	sapi_mutex_unlock(&__mutex);

	if(ret != OK) {
		CTRANS_DEBUG_PRINTF("ctrans_req(): sapi_hash_ins(): failed!!!!!");
		ct_deref(ct);
	}
	else {
		ctrans_tx(ct, ct->req);

		if(ct->tp->proto == UDP)
			sapi_oneshot(T1, retrans, NULL, ct->hkey);

		sapi_oneshot(T1 * 64, timeout, NULL, ct->hkey);

		/* CANCEL Handling */
		/* @todo don't send cancel if oct state is not PROCEEDING */
		if(sapi_strcmp(req->request->method, "CANCEL")) {

			u8_t *ohkey = ctrans_hash_key(req, TRUE);
			ct_t *oct;

			if(ohkey) {
				sapi_mutex_lock(&__mutex);
				oct = sapi_hash_find(__ht, ohkey);
				sapi_mutex_unlock(&__mutex);

				if(oct) {
					CTRANS_DEBUG_PRINTF("ctrans_req(): Making sure original INVITE times out");
					sapi_oneshot(T1 * 64, ctrans_timeout_cancel_i, NULL, oct->hkey);
				}
				sapi_free(ohkey);
			}
		}
	}

	ct_deref(ct);
}


ret_t ctrans_resp(const msg_t *resp)
{
	u8_t *hkey = NULL;
	ct_t *ct;
	ret_t ret = ERR;

	if(!__ht || !resp || !resp->response) {
		CTRANS_DEBUG_PRINTF("ctrans_resp(): __ht, resp or resp->response is NULL");
		goto out;
	}

	hkey = ctrans_hash_key(resp, FALSE);
	if(!hkey) {
		CTRANS_DEBUG_PRINTF("ctrans_resp(): hkey is NULL");
		goto out;
	}

	sapi_mutex_lock(&__mutex);
	ct = ct_ref(sapi_hash_find(__ht, hkey));
	sapi_mutex_unlock(&__mutex);

	if(!ct) {
		CTRANS_DEBUG_PRINTF("ctrans_resp(): ctrans not found");
		goto out;
	}
	else {
		if(sapi_strcmp(ct->req->request->method, "INVITE") == TRUE) {
			ctrans_resp_i(ct, resp);
		}
		else {
			ctrans_resp_ni(ct, resp);
		}

		ct_deref(ct);
	}

	ret = OK;

out:
	sapi_free(hkey);
	return(ret);
}


void ctrans_status(void)
{
	if(!__ht)
		return;

	sapi_mutex_lock(&__mutex);
	sapi_hash_status(__ht);
	sapi_mutex_unlock(&__mutex);
}


u8_t *ctrans_status_str(void)
{
	u8_t *str;

	if(!__ht)
		return("ctrans_status_str(): __ht is NULL");

	sapi_mutex_lock(&__mutex);
	str = sapi_hash_status_str(__ht);
	sapi_mutex_unlock(&__mutex);

	return(str);
}
