/**
 * @file g711.c Implements G.711 codec.
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>.  All rights reserved.
 */
#include <sapi.h>
#include <g711.h>


#define G711_NSEGS      8        /**< Number of A-law segments. */
#define	G711_QUANT_MASK	0x0f     /**< Quantization field mask. */
#define	G711_SEG_SHIFT  4        /**< Left shift for segment number. */
#define	G711_SEG_MASK   0x70     /**< Segment field mask. */
#define	G711_SIGN_BIT   0x80     /**< Sign bit for a A-law byte. */
#define G711_CLIP       32635    /**< Maximum absolute value for ulaw PCM input. */


/**
 * Determine which segment the PCM sample matches.
 *
 * @param val Absolute magnitude of PCM sample.
 *
 * @return 3-bit segment.
 */
static u8_t seg_search(s16_t val)
{
	static const s16_t alaw_segs[G711_NSEGS] =
		{0xff, 0x1ff, 0x3ff, 0x7ff, 0xfff, 0x1fff, 0x3fff, 0x7fff};
	u8_t i;

	for(i=0; i<G711_NSEGS; i++) {
		if(val <= alaw_segs[i])
			return(i);
	}
	return(G711_NSEGS);
}


/**
 * Convert from PCM to Alaw.
 *
 * @param pcm Signed 16-bit PCM input value.
 *
 * @return 8-bit Alaw.
 */
u8_t pcm2alaw(s16_t pcm)
{
	u8_t mask, seg, alaw;

	if(pcm >= 0) {			/* sign (7th) bit = 1 */
		mask = 0xd5;		/* 1101_0101*/

	} else {			/* sign bit = 0 */
		mask = 0x55;		/* 0101_0101*/
		pcm = -pcm - 8;
	}

	/* Convert the scaled magnitude to segment number. */
	seg = seg_search(pcm);

	/* Combine the sign, segment, and quantization bits. */
	if(seg >= G711_NSEGS)		/* out of range, return maximum value. */
		return(0x7f ^ mask);
	else {
		alaw = seg << G711_SEG_SHIFT;
		if(seg < 2)
			alaw |= (pcm >> 4) & G711_QUANT_MASK;	/* TODO >> 5 ? */
		else
			alaw |= (pcm >> (seg + 3)) & G711_QUANT_MASK;	/* TODO >> (seg + 4) ?  */

		return(alaw ^ mask);
	}
}


/**
 * Convert from Alaw to PCM.
 *
 * @param alaw 8-bit Alaw input value.
 *
 * @return Signed 16-bit PCM value.
 */
s16_t alaw2pcm(u8_t alaw)
{
	s16_t pcm;
	u8_t seg;

	alaw ^= 0x55;

	pcm = (alaw & G711_QUANT_MASK) << 4;
	seg = (alaw & G711_SEG_MASK) >> G711_SEG_SHIFT;
	switch (seg) {
	case 0:
		pcm += 8;
		break;
	case 1:
		pcm += 0x108;
		break;
	default:
		pcm += 0x108;
		pcm <<= seg - 1;
		break;
	}
	return((alaw & G711_SIGN_BIT) ? pcm : -pcm);
}


/**
 * Convert from PCM to Mulaw.
 *
 * @param sample Signed 16-bit PCM input value.
 *
 * @return 8-bit Mulaw.
 */
u8_t pcm2ulaw(s16_t sample)
{
 	u8_t ulawbyte, sign, mantissa, exponent;  /* Get the sample into sign-magnitude. */

	if(sample >= 0) {
		sign = 0;
	} else {
		sign = 0x80;
		sample = -sample;
	}

	/* clip the magnitude */
	if(sample > G711_CLIP)
		sample = G711_CLIP;

 	/* Convert from 16 bit linear to ulaw. */

 	/* Convert the scaled magnitude to segment number. */
	exponent = seg_search(sample);
 	mantissa = (sample >> (exponent + 3)) & 0x0f;
 	ulawbyte = ~(sign | (exponent << 4) | mantissa);

	return(ulawbyte);
}


/**
 * Convert from Mulaw to PCM.
 *
 * @param ulawbyte 8-bit Mulaw input value.
 *
 * @return Signed 16-bit PCM value.
 */
s16_t ulaw2pcm(u8_t ulawbyte)
{
	static u16_t exp_lut[8] = {0,132,396,924,1980,4092,8316,16764};

	u16_t sign, exponent, mantissa;
	s16_t sample;

	ulawbyte = ~ulawbyte;
	sign = (ulawbyte & 0x80);
	exponent = (ulawbyte >> 4) & 0x07;
	mantissa = ulawbyte & 0x0f;
	sample = exp_lut[exponent] + (mantissa << (exponent + 3));
	if(sign != 0) sample = -sample;
	return(sample);
}
