/**
 * @file msg.h Interface to msg submodule
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>.  All rights reserved.
 */
#ifndef __MSG_H__
#define __MSG_H__


#define MSG_DEBUG 1

#ifdef MSG_DEBUG
#define MSG_DEBUG_PRINTF sapi_logf
#else
#define MSG_DEBUG_PRINTF(a, b...)
#endif


#define SIP_VERSION "SIP/2.0"
#define SIP_PORT (5060)


/* param_t */
typedef struct {
	u8_t *name;
	u8_t *value;
	node_t *node;
} param_t;


/* url_t */
typedef struct {
	u8_t *scheme;
	u8_t *username;
	u8_t *password;
	u8_t *host;
	u16_t port;
	list_t *params;
	list_t *headers;
} url_t;


/* address_t */
typedef struct {
	u8_t *displayname;
	url_t *url;
	list_t *params;
} address_t;


/* cseq_t */
typedef struct {
	u32_t num;
	u8_t *method;
} cseq_t;


/* via_t */
typedef struct {
	u8_t *ver;
	u8_t *proto;
	u8_t *host;
	u16_t port;
	list_t *params;
} via_t;


/* request_t */
typedef struct {
	u8_t *method;
	url_t *url;
	u8_t *ver;
} request_t;


/* response_t */
typedef struct {
	u8_t *ver;
	u8_t *statuscode;
	u8_t *reasonphrase;
} response_t;


/* msg_t */
typedef struct _msg_t {
	request_t *request;       ///< If non-zero, contains the SIP request.
	response_t *response;     ///< If non-zero, contains the SIP response.
	list_t *headers;          ///< List of unparsed headers in SIP message.
	list_t *vias;             ///< List of Via fields in SIP message.
	address_t *to;            ///< Contains the To: address in the SIP message.
	address_t *from;          ///< Contains the From: address in the SIP message.
	list_t *contacts;         ///< List of Contact fields in the SIP message.
	cseq_t *cseq;             ///< Contains the CSeq method and number.
	u8_t *call_id;            ///< Contains the CallID.
	u32_t *max_forwards;      ///< Contains the Max-Forwards header field.
	u32_t *expires;           ///< Contains the Expires header field.
	u8_t *user_agent;         ///< Contains the value part of the user-agent header field
	u8_t *ssid;               ///< Contains the SSID for Wireless LAN phones.
	u8_t *body;               ///< If non-zero, contains the body (e.g. SDP)
	u32_t nrefs;              ///< Reference count for "automatic" garbage collection
	mutex_t mutex;
} msg_t;


/* msg_param */
param_t *msg_param_alloc();
void msg_param_free(void *ptr);
param_t *msg_param_dup(const param_t *param);


/* msg_param_list */
list_t *msg_param_list_dup(const list_t *params);
void msg_param_list_add_param(list_t *params, const u8_t *name, const u8_t *value);
void msg_param_list_add_param_u32(list_t *params, const u8_t *name, u32_t value);
u8_t *msg_param_list_get_param(const list_t *params, const u8_t *name);
ret_t msg_param_list_parse(list_t *params, const u8_t *start, const u8_t *end, u8_t sep_sign);
u8_t *msg_param_list_str(const list_t *params, u8_t sep_sign);


/* msg_url */
url_t *msg_url_alloc();
void msg_url_free(url_t *url);
url_t *msg_url_dup(const url_t *url);
ret_t msg_url_parse(url_t *url, const u8_t *url_str);
u8_t *msg_url_str(const url_t *url);
u8_t *msg_url_cannonical(const url_t *url);


/* msg_address */
address_t *msg_address_alloc();
void msg_address_free(void *ptr);
address_t *msg_address_dup(const address_t *address);
void msg_address_addparam(address_t *address, const u8_t *name, const u8_t *value);
void msg_address_addparam_u32(address_t *address, const u8_t *name, u32_t value);
u8_t *msg_address_getparam(const address_t *address, const u8_t *name);
ret_t msg_address_parse(address_t *address, const u8_t *address_str);
u8_t *msg_address_str(const address_t *address);
u8_t *msg_address_display_str(const address_t *address);


/* cseq */
cseq_t *msg_cseq_alloc();
void msg_cseq_free(cseq_t *cseq);
ret_t msg_cseq_parse(cseq_t *cseq, const u8_t *cseq_str);
u8_t *msg_cseq_str(const cseq_t *cseq);


/* via */
via_t *msg_via_alloc();
void msg_via_free(void *ptr);
via_t *msg_via_dup(const via_t *via);
via_t *msg_via_create(const u8_t *host, u16_t port);
void msg_via_addparam(via_t *via, const u8_t *name, const u8_t *value);
u8_t *msg_via_getparam(const via_t *via, const u8_t *name);
ret_t msg_via_parse(via_t *via, const u8_t *via_str);
u8_t *msg_via_str(const via_t *via);


/* request */
request_t *msg_request_alloc(void);
void msg_request_free(request_t *request);
ret_t msg_request_parse(request_t *request, const u8_t *request_str);
u8_t *msg_request_str(const request_t *request);


/* response */
response_t *msg_response_alloc(void);
void msg_response_free(response_t *response);
ret_t msg_response_parse(response_t *response, const u8_t *response_str);
u8_t *msg_response_str(const response_t *response);


/* msg */
msg_t *msg_alloc(void);
const msg_t *msg_ref(const msg_t *msg);
void msg_deref(const msg_t *msg);
void msg_vderef(void *ptr);
msg_t *msg_dup(const msg_t *msg);
ret_t msg_parse(msg_t *msg, const u8_t *payload);
u8_t *msg_str(const msg_t *msg);
void msg_set_request(msg_t *msg, const u8_t *method, const url_t *url);
void msg_set_response(msg_t *msg, const u8_t *statuscode, const u8_t *reasonphrase);
void msg_add_via(msg_t *msg, const via_t *via);
void msg_prepend_via(msg_t *msg, const via_t *via);
void msg_set_to(msg_t *msg, const address_t *to);
void msg_set_from(msg_t *msg, const address_t *from);
void msg_add_contact(msg_t *msg, const address_t *contact);
void msg_set_call_id(msg_t *msg, const u8_t *call_id);
void msg_set_cseq(msg_t *msg, u32_t num, const u8_t *method);
void msg_set_max_forwards(msg_t *msg, u32_t max_forwards);
void msg_dec_max_forwards(msg_t *msg);
void msg_set_expires(msg_t *msg, u32_t expires);
void msg_set_user_agent(msg_t *msg, const u8_t *user_agent);
void msg_set_ssid(msg_t *msg, const u8_t *ssid);
void msg_add_header(msg_t *msg, const u8_t *hname, const u8_t *hvalue);
void msg_add_header_u32(msg_t *msg, const u8_t *hname, u32_t hvalue);
void msg_set_body(msg_t *msg, const u8_t *body);
via_t *msg_get_topvia(const msg_t *msg);
void msg_del_topvia(msg_t *msg);
u32_t msg_nvias(const msg_t *msg);
msg_t *msg_response(const msg_t *req, const u8_t *statuscode, const u8_t *reasonphrase);
msg_t *msg_cancel(const msg_t *req);
msg_t *msg_ack(const msg_t *req, const msg_t *resp);
void msg_dump(const msg_t *msg);


#endif /* __MSG_H__ */
