#ifndef __SAPI_H__
#define __SAPI_H__


#define SAPI_DEBUG 1

#ifdef SAPI_DEBUG
#define SAPI_DEBUG_PRINTF sapi_logf
#else
#define SAPI_DEBUG_PRINTF(a, b...)
#endif


#define DEBUG(a, b...) sapi_logf("#DEBUG: " __FUNCTION__ "(): " a, ## b)
#define ERROR(a, b...) sapi_logf("#ERROR: " __FUNCTION__ "(): " a, ## b)
#define LOG(a, b...) sapi_logf(a, ## b)


#include <pthread.h> // @todo


/*
#define SAPI_LOCK(a) sapi_logf("*** Acquring lock @ " __FILE__ ":%i", __LINE__); \
                     pthread_mutex_lock(a); \
                     sapi_logf("*** Lock acquired @ " __FILE__ ":%i", __LINE__);

#define SAPI_UNLOCK(a) sapi_logf("*** Releasing lock @ " __FILE__ ":%i", __LINE__); \
                       pthread_mutex_unlock(a); \
                       sapi_logf("*** Lock released @ " __FILE__ ":%i", __LINE__);
*/

#define SAPI_LOCK(a) pthread_mutex_lock(a);

#define SAPI_UNLOCK(a) pthread_mutex_unlock(a);





#ifndef NULL
#define NULL ((void*)0)
#endif


/* SAPI types */
typedef unsigned char u8_t;
typedef signed char s8_t;
typedef unsigned short int u16_t;
typedef signed short int s16_t;
typedef unsigned int u32_t;
typedef signed int s32_t;

typedef enum _ret_t {
	     ERR = -1,
	     OK = 0
} ret_t;

typedef enum _bool_t {
	FALSE = 0,
	TRUE = 1
} bool_t;

typedef pthread_mutex_t mutex_t;


/* Generic free function */
typedef void(*free_t)(void *ptr);

typedef void(*sapicb_udp_listen)(u8_t *str);
typedef void(*sapicb_log)(u8_t *str);


#define SAPI_READ  0
#define SAPI_WRITE 1


/* SAPI functions */
void sapi_printf(const u8_t *fmt, ...);
void sapi_printf2(const u8_t *fmt, ...);

void sapi_log_init(const u8_t *logfile);
void sapi_log_init2(sapicb_log cb);
void sapi_logf(const u8_t *fmt, ...);
void sapi_log_status(void);
u8_t *sapi_log_status_str(void);

u8_t *sapi_sprintf(const u8_t *fmt, ...);
u8_t *sapi_saprintf(u8_t **astr, const u8_t *fmt, ...);
void *sapi_malloc(u32_t size);
void *sapi_realloc(void *ptr, u32_t size);
void sapi_free(void *ptr);
u32_t sapi_nmalloced(void);
void sapi_bluesmoke(const u8_t *msg);
u32_t sapi_strlen(const u8_t *str);
bool_t sapi_strcmp(const u8_t *str1, const u8_t *str2);
bool_t sapi_strcmp_ci(const u8_t *str1, const u8_t *str2);
bool_t sapi_strncmp(const u8_t *str1, const u8_t *str2, u32_t n);
u8_t *sapi_strchr(const u8_t *str, u8_t chr);
u8_t *sapi_strdup(const u8_t *str);
u8_t *sapi_strndup(const u8_t *str, u32_t len);
bool_t sapi_is_numeric(const u8_t *str);
u32_t sapi_to_number(const u8_t *start, const u8_t *end);
u32_t sapi_to_number2(const u8_t *str);
ret_t sapi_to_number3(u32_t *num, const u8_t *str, u32_t strlen);
u8_t *sapi_lower(u8_t *str);
u32_t sapi_min(u32_t a, u32_t b);

void sapi_daemon(void);
void sapi_udp_listen(u16_t port, sapicb_udp_listen cb);
void sapi_udp_listen2(u32_t *fd, u16_t port);
void sapi_udp_send(const u8_t *host, const u16_t port, const void *payload, u32_t len);
ret_t sapi_tcp_listen(u32_t *fd, u16_t port);
ret_t sapi_resolv(const u8_t *hostname, u32_t *hostaddr);

ret_t sapi_open(u32_t *fd, const u8_t *pathname, u8_t mode);
ret_t sapi_close(u32_t fd);
ret_t sapi_unlink(const u8_t *pathname);
ret_t sapi_fprintf(u32_t fd, const u8_t *fmt, ...);
u8_t *sapi_freadline(u32_t fd);




/* List API */
typedef struct _node_t {
	void *value;
	struct _node_t *next;
	struct _node_t *prev;
} node_t;

typedef struct _list_t {
	node_t *start;
	node_t *end;
	u32_t n;
	u32_t tot;
	u32_t max;
} list_t;

// typedef void(*list_free_t)(void *ptr);
typedef ret_t(*list_apply_t)(void *value, u32_t n, void *arg1, void *arg2);
typedef bool_t(*list_find_t)(void *value, void *arg);


list_t *sapi_list_alloc(void);
void sapi_list_free(list_t *list, free_t free);
node_t *sapi_list_prepend(list_t *list, void *value);
node_t *sapi_list_append(list_t *list, void *value);
void sapi_list_remove(list_t *list, node_t *node, free_t free);
ret_t sapi_list_apply(list_t *list, list_apply_t apply, void *arg1, void *arg2);
void *sapi_list_find(list_t *list, list_find_t find, void *arg);
void *sapi_list_get_head(const list_t *list);
void sapi_list_del_head(list_t *list, free_t free);
u32_t sapi_list_nnodes(const list_t *list);
void sapi_list_status(const list_t *list);
u8_t *sapi_list_status_str(const list_t *list);
void sapi_list_dump(const list_t *list);




/* Hashtable API */
typedef void(*hash_free_t)(void *ptr);


typedef struct {
	u32_t size;
	list_t **buckets;
	hash_free_t free;
	u32_t n;
	u32_t tot;
	u32_t max;
	u32_t bucket_max;
} hash_t;

typedef ret_t(*hash_apply_t)(void *value, void *arg);

hash_t *sapi_hash_alloc(u32_t size);
void sapi_hash_free(hash_t *ht);
ret_t sapi_hash_ins(hash_t *ht, const u8_t *key, void *value, hash_free_t free);
void *sapi_hash_find(const hash_t *ht, const u8_t *key);
void sapi_hash_del(hash_t *ht, const u8_t *key);
ret_t sapi_hash_apply(hash_t *ht, hash_apply_t apply, void *arg);

void sapi_hash_status(const hash_t *ht);
u8_t *sapi_hash_status_str(const hash_t *ht);
void sapi_hash_dump(const hash_t *ht);
void sapi_hash_histogram(const hash_t *ht);


/* FIFO API */
typedef struct {
	node_t *head;
	node_t *tail;
	pthread_mutex_t mutex;
	pthread_cond_t cond;
	u32_t n;
	u32_t tot;
	u32_t max;
} fifo_t;

typedef void(*fifo_free_t)(void *ptr);


fifo_t *sapi_fifo_alloc(void);
void sapi_fifo_free(fifo_t *fifo, fifo_free_t freefunc);
void sapi_fifo_put(fifo_t *fifo, const void *value);
void *sapi_fifo_get(fifo_t *fifo);
u32_t sapi_fifo_nnodes(fifo_t *fifo);
void sapi_fifo_status(fifo_t *fifo);
u8_t *sapi_fifo_status_str(fifo_t *fifo);



/* Oneshot API */

typedef void(*timeout_t)(void *arg, u8_t *sarg);

void sapi_oneshot_init(void);
void sapi_oneshot(u32_t time, timeout_t timeout, void *arg, const u8_t *sarg);
void sapi_oneshot_status(void);
u8_t *sapi_oneshot_status_str(void);





/* Mutex API */

#define sapi_mutex_lock   pthread_mutex_lock
#define sapi_mutex_unlock pthread_mutex_unlock
#define SAPI_MUTEX_INITIALIZER  PTHREAD_MUTEX_INITIALIZER

void sapi_mutex_init(mutex_t *mutex);




/*
 * SymbianOS specific types
 */
#if defined(__SYMBIAN32__)
#ifndef socklen_t
#define socklen_t int
#endif
#endif /* __SYMBIAN32__ */


#define sapi_bzero(a,b) memset(a, 0, b)

#undef FD_ZERO
#define	FD_ZERO(p)	sapi_bzero((caddr_t)(p), sizeof (*(p)))

#define WARNING sapi_printf
#define ASSERT(a) {if(!a) sapi_printf("ASSERT: %s:%d", __FILE__, __LINE__);}

#endif /* __SAPI_H__ */
