/**
 * @file sdp.h Defines the interface to the SDP module.
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>.  All rights reserved.
 */
#ifndef SDP_H__
#define SDP_H__


/**
 * This struct contains one media attribute. The rtpmap and fmtp elements
 * are for dynamic payload types and they are optional.
 *
 * Example:
 * <pre>
 * a=rtpmap:0 PCMU/8000
 * </pre>
 */
typedef struct {
	u8_t rtp_pt;      /**< RTP payload type as of defined in RFC 1890. */
	u8_t *rtpmap;     /**< RTP codec string e.g. "PCMU". */
	u16_t clockrate;  /**< Clockrate e.g. 8000. */
	u8_t *fmtp;       /**< Format list with e.g. bitrates. */
} sdp_attrib_t;


/**
 * This struct contains one media type from the SDP message.
 *
 * Example:
 * <pre>
 * m=audio 5004 RTP/AVP 4 0
 * </pre>
 */
typedef struct {
	u8_t *type;       /**< Media type e.g. "audio" or "video". */
	u16_t port;       /**< Port from SDP fields. */
	u8_t *transp;     /**< Transport type e.g. "RTP/AVP". */
	list_t *attribs;  /**< Linked list of media attributes. */
} sdp_media_t;


/**
 * This struct holds information from the SDP body.
 */
typedef struct {
	u8_t *host;          /**< Internet address from SDP fields. */
	sdp_media_t *media;  /**< Pointer to the media type object. */
} sdp_t;


/*
 * Prototypes
 */
sdp_t *sdp_create(const u8_t *host, const u8_t *media_type, u16_t media_port, const u8_t *media_transp);
void sdp_media_add(sdp_t *sdp, u8_t rtp_pt, const u8_t *rtpmap, u16_t clockrate, const u8_t *fmtp);
u8_t *sdp_str(const sdp_t *sdp);
sdp_t *sdp_parse(const u8_t *body);
void sdp_free(sdp_t *sdp);
void attrib_free(void *ptr);
void sdp_dump(const sdp_t *sdp);

#endif /* SDP_H__ */
