#include <time.h>
#include <sapi.h>
#include <msg.h>
#include <ldb.h>


#define LDB_HASH_SIZE (32768)
#define DEFAULT_EXPIRES (3600)


typedef struct {
	u8_t *address;
	list_t *bindings;
} entry_t;


typedef struct {
	u8_t *location;
	u8_t *call_id;
	u32_t cseq_num;
	u8_t *user_agent;
	u8_t *ssid;
	u32_t expires;
	entry_t *entry;
	node_t *node;
	u32_t nshot;
	u32_t epoc;
} binding_t;


typedef struct {
	u8_t *address;
	u8_t *location;
} addrloc_t;


static hash_t *__ht = NULL;
static pthread_mutex_t __mutex = PTHREAD_MUTEX_INITIALIZER;


static inline addrloc_t *addrloc_alloc()
{
	return((addrloc_t *)sapi_malloc((u32_t)sizeof(addrloc_t)));
}


static void addrloc_free(addrloc_t *al)
{
	if(!al) {
		LDB_DEBUG_PRINTF("addrloc_free(): al is NULL");
		return;
	}

	sapi_free(al->address);
	sapi_free(al->location);
	sapi_free(al);
}


static inline binding_t *binding_alloc()
{
	return((binding_t *)sapi_malloc((u32_t)sizeof(binding_t)));
}


static void binding_free(void *ptr)
{
	binding_t *bnd = (binding_t *)ptr;

	if(!bnd) {
		LDB_DEBUG_PRINTF("binding_free(): bnd is NULL");
		return;
	}

	sapi_free(bnd->location);
	sapi_free(bnd->call_id);
	sapi_free(bnd->user_agent);
	sapi_free(bnd->ssid);
	sapi_free(bnd);
}


static inline entry_t *entry_alloc()
{
	return((entry_t *)sapi_malloc((u32_t)sizeof(entry_t)));
}


static void entry_free(void *ptr)
{
	entry_t *entry = (entry_t *)ptr;

	if(!entry) {
		LDB_DEBUG_PRINTF("entry_free(): entry is NULL");
		return;
	}

	sapi_free(entry->address);
	sapi_list_free(entry->bindings, binding_free);
	sapi_free(entry);
}


static bool_t binding_find(void *value, void *arg)
{
	binding_t *bnd = (binding_t *)value;
	u8_t *location = (u8_t *)arg;

	if(!bnd || !location) {
		LDB_DEBUG_PRINTF("binding_find(): bnd or location is NULL");
		return(ERR);
	}

	return(sapi_strcmp(bnd->location, location));
}


static void binding_timeout(void *arg, u8_t *sarg)
{
	addrloc_t *al = (addrloc_t *)arg;
	entry_t *entry;
	binding_t *bnd;

	if(!al) {
		LDB_DEBUG_PRINTF("binding_timeout(): al is NULL");
		return;
	}

	SAPI_LOCK(&__mutex);

	entry = (entry_t *)sapi_hash_find(__ht, al->address);
	if(!entry) {
		LDB_DEBUG_PRINTF("binding_timeout(): entry not found");
		goto out;
	}

	bnd = (binding_t *)sapi_list_find(entry->bindings, binding_find, al->location);
	if(!bnd) {
		LDB_DEBUG_PRINTF("binding_timeout(): binding not found");
		goto out;
	}

	if((--bnd->nshot)) {
		goto out;
	}

	LDB_DEBUG_PRINTF("binding_timeout(): expiring binding '%s' for address '%s'",
			 bnd->location, entry->address);

	sapi_list_remove(entry->bindings, bnd->node, binding_free);


	/* Remove entry if zero bindings */
	if(!sapi_list_nnodes(entry->bindings))
		sapi_hash_del(__ht, al->address);

out:
	addrloc_free(al);
	SAPI_UNLOCK(&__mutex);
}


static inline u32_t expires(const address_t *contact, const msg_t *msg)
{
	u32_t exp     = DEFAULT_EXPIRES;
	u8_t *exp_str = msg_address_getparam(contact, "expires");

	if(!exp_str && msg->expires) {
		exp = *(msg->expires);
	}
	else if(sapi_to_number3(&exp, exp_str, sapi_strlen(exp_str)) != OK) {
		LDB_DEBUG_PRINTF("binding_put(): expires contact parameter is not numeric");
	}

	return(exp);
}


static ret_t binding_put(void *value, u32_t n, void *arg1, void *arg2)
{
	address_t *contact = (address_t *)value;
	entry_t *entry = (entry_t *)arg1;
	msg_t *msg = (msg_t *)arg2;
	binding_t *bnd;
	addrloc_t *al;
	u8_t *location;
	u32_t exp;

	if(!contact || !entry || !msg) {
		LDB_DEBUG_PRINTF("binding_put(): contact, entry or msg is NULL");
		return(ERR);
	}

	exp = expires(contact, msg);

	location = sapi_lower(msg_url_cannonical(contact->url));
	bnd = (binding_t *)sapi_list_find(entry->bindings, binding_find, location);
	sapi_free(location);

	if(!bnd) {
		/* Insert binding */
		bnd = binding_alloc();

		bnd->location = sapi_lower(msg_url_cannonical(contact->url));
		bnd->call_id  = sapi_strdup(msg->call_id);
		bnd->cseq_num = msg->cseq->num;

		if(msg->user_agent)
			bnd->user_agent  = sapi_strdup(msg->user_agent);

		if(msg->ssid)
			bnd->ssid  = sapi_strdup(msg->ssid);

		bnd->expires  = exp;
		bnd->entry    = entry;
		bnd->node     = sapi_list_append(entry->bindings, bnd);
		bnd->nshot    = 1;
		bnd->epoc     = (u32_t)time(NULL);

		al  = addrloc_alloc();

		al->address   = sapi_strdup(entry->address);
		al->location  = sapi_strdup(bnd->location);

		sapi_oneshot(exp * 10, binding_timeout, al, NULL);

		LDB_DEBUG_PRINTF("ldb_put(): Creating new binding '%s' for entry '%s', expires: %u",
				 bnd->location, entry->address, bnd->expires);
	}
	else {
		if(sapi_strcmp(bnd->call_id, msg->call_id) && msg->cseq->num <= bnd->cseq_num) {

			/* Ignoring */
			LDB_DEBUG_PRINTF("ldb_put(): Ignoring bind update (CSeq/Call-ID check)");
		}
		else if(exp) {

			/* Updating Binding */
			if(bnd->call_id)
				sapi_free(bnd->call_id);

			bnd->call_id  = sapi_strdup(msg->call_id);
			bnd->cseq_num = msg->cseq->num;

			/* User-Agent */
			if(bnd->user_agent){
				sapi_free(bnd->user_agent);
				bnd->user_agent = NULL;
			}

			if(msg->user_agent)
				bnd->user_agent = sapi_strdup(msg->user_agent);

			/* SSID */
			if(bnd->ssid){
				sapi_free(bnd->ssid);
				bnd->ssid = NULL;
			}

			if(msg->ssid)
				bnd->ssid = sapi_strdup(msg->ssid);

			bnd->expires  = exp;
			bnd->nshot++;
			bnd->epoc     = (u32_t)time(NULL);

			al  = addrloc_alloc();

			al->address   = sapi_strdup(entry->address);
			al->location  = sapi_strdup(bnd->location);

			sapi_oneshot(exp * 10, binding_timeout, al, NULL);

			LDB_DEBUG_PRINTF("ldb_put(): Updating binding '%s' for entry '%s', expires: %u",
					 bnd->location, entry->address, bnd->expires);
		}
		else {
			LDB_DEBUG_PRINTF("ldb_put(): Removing binding '%s' for entry '%s'",
					 bnd->location, entry->address);

			sapi_list_remove(entry->bindings, bnd->node, binding_free);

			/* Remove entry if zero bindings */
			if(!sapi_list_nnodes(entry->bindings))
				sapi_hash_del(__ht, entry->address);
		}
	}

	return(OK);
}


void ldb_init(void)
{
	if(__ht)
		return;

	__ht = sapi_hash_alloc(LDB_HASH_SIZE);

	sapi_logf("Location database service is up and running");
}


ret_t ldb_put(const msg_t *msg)
{
	entry_t *entry;
	address_t *contact;
	u8_t *addr;


	if(!__ht || !msg || !msg->contacts || !sapi_list_nnodes(msg->contacts)) {
		LDB_DEBUG_PRINTF("ldb_put(): __ht, msg or msg->contacts is NULL");
		return(ERR);
	}

	SAPI_LOCK(&__mutex);

	addr    = msg_url_cannonical(msg->to->url);
	entry   = sapi_hash_find(__ht, addr);
	contact = (address_t *)sapi_list_get_head(msg->contacts);


	/* Remove all bindings; 'Contact: *' */
	if(entry && contact && sapi_strcmp(contact->displayname, "*")) {

		LDB_DEBUG_PRINTF("ldb_put(): Removing entry '%s'", addr);
		sapi_hash_del(__ht, addr);
	}
	else {
		if(!entry) {

			entry = entry_alloc();

			entry->address  = sapi_lower(sapi_strdup(addr));
			entry->bindings = sapi_list_alloc();

			sapi_hash_ins(__ht, addr, entry, entry_free);

			LDB_DEBUG_PRINTF("ldb_put(): Creating new entry for '%s'", addr);
		}

		sapi_list_apply(msg->contacts, binding_put, entry, (void *)msg);
	}

	sapi_free(addr);

	SAPI_UNLOCK(&__mutex);

	return(OK);
}


static ret_t build_contacts(void *value, u32_t n, void *arg1, void *arg2)
{
	binding_t *bnd = (binding_t *)value;
	list_t *contacts = (list_t *)arg1;
	address_t *contact;
	u32_t left;


	if(!bnd || !contacts) {
		LDB_DEBUG_PRINTF("LDB: build_contacts(): bnd or contacts is NULL");
		return(ERR);
	}

	contact = msg_address_alloc();

	if(msg_address_parse(contact, bnd->location) != OK) {
		LDB_DEBUG_PRINTF("LDB: build_contacts(): error parsing contact");
		msg_address_free(contact);
		return(ERR);
	}

	left = bnd->expires + bnd->epoc - time(NULL);
	if(left < 0)
		left = 0;

	msg_address_addparam_u32(contact, "expires", left);

	sapi_list_append(contacts, contact);

	return(OK);
}


list_t *ldb_get(const u8_t *address)
{
	entry_t *entry;
	list_t *contacts = NULL;

	SAPI_LOCK(&__mutex);

	entry   = sapi_hash_find(__ht, address);
	if(!entry || !sapi_list_nnodes(entry->bindings)) {
		LDB_DEBUG_PRINTF("ldb_get(): Can't find any bindings for '%s'", address);
		goto out;
	}

	contacts = sapi_list_alloc();

	if(sapi_list_apply(entry->bindings, build_contacts, contacts, NULL) != OK) {
		LDB_DEBUG_PRINTF("ldb_get(): Can't build contact list");
		sapi_list_free(contacts, msg_address_free);
		contacts = NULL;
		goto out;
	}

out:
	SAPI_UNLOCK(&__mutex);

	return(contacts);
}


void ldb_status(void)
{
	if(!__ht)
		return;

	SAPI_LOCK(&__mutex);
	sapi_hash_status(__ht);
	SAPI_UNLOCK(&__mutex);
}


static ret_t dump_binding(void *value, u32_t n, void *arg1, void *arg2)
{
       binding_t *bnd = (binding_t *)value;
       u8_t **str = (u8_t **)arg1;
       s32_t left;

       if(!bnd || !str) {
	       LDB_DEBUG_PRINTF("LDB: dump_binding(): bnd or str is NULL");
	       return(ERR);
       }

       left = bnd->expires + bnd->epoc - time(NULL);
       if(left < 0)
	       left = 0;

       if(n)
	       sapi_saprintf(str, "   <tr bgcolor=\"#ffffff\">\n");

       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%s</font></td>\n", bnd->location);
       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%u/%u sec(s)</font></td>\n",
		     left, bnd->expires);
       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%s</font></td>\n", bnd->call_id);
       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%u</font></td>\n", bnd->cseq_num);

       /* User-Agent */
       if(bnd->user_agent)
	       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%s</font></td>\n", bnd->user_agent);
       else
	       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">&nbsp;</font></td>\n");

       /* SSID */
       if(bnd->ssid)
	       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%s</font></td>\n", bnd->ssid);
       else
	       sapi_saprintf(str, "    <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">&nbsp;</font></td>\n");

       sapi_saprintf(str, "   </tr>\n");

       return(OK);
}


static ret_t dump_entry(void *value, void *arg)
{
	entry_t *entry = (entry_t *)value;
	u8_t **str = (u8_t **)arg;

	if(!entry || !str) {
		LDB_DEBUG_PRINTF("LDB: dump_entry(): entry or str is NULL");
		return(ERR);
	}

	sapi_saprintf(str, "   <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(str, "    <td rowspan=\"%u\" valign=\"top\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\"><b>%s</b></font></td>\n",
		      sapi_list_nnodes(entry->bindings), entry->address);

	sapi_list_apply(entry->bindings, dump_binding, str, NULL);

	sapi_saprintf(str, " </tr>\n");

	return(OK);
}


u8_t *ldb_dump_html(void)
{
	u8_t *str = NULL;

	if(!__ht)
		return(NULL);

	sapi_saprintf(&str, "  <table bgcolor=\"#ffffff\" border=\"0\" cellpadding=\"0\" cellspacing=\"0\" width=\"100%\">\n");
	sapi_saprintf(&str, "   <tr bgcolor=\"1478EB\">\n");
	sapi_saprintf(&str, "    <td colspan=\"7\"><font face=\"Arial,Helvetica\" color=\"#ffffff\"><b>Active Registrations</b></font></td>\n");
	sapi_saprintf(&str, "   </tr>\n");
	sapi_saprintf(&str, "   <tr bgcolor=\"1478EB\">\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>Address</b></font></td>\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>Location</b></font></td>\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>Expires</b></font></td>\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>Call-ID</b></font></td>\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>CSeq</b></font></td>\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>User-Agent</b></font></td>\n");
	sapi_saprintf(&str, "    <td ><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>SSID</b></font></td>\n");
	sapi_saprintf(&str, "   </tr>\n");

	SAPI_LOCK(&__mutex);
	sapi_hash_apply(__ht, dump_entry, &str);
	SAPI_UNLOCK(&__mutex);

	sapi_saprintf(&str, "</table>\n");

	return(str);
}
