/**
 * @file msg.c Implements code for handle SIP messages
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>. All rights reserved.
 */
#include <sapi.h>
#include <msg.h>


static ret_t via_dup(void *value, u32_t n, void *arg1, void *arg2)
{
	via_t *via  = (via_t *)value;
	msg_t *msg = (msg_t *)arg1;

	if(!via || !msg)
		return(ERR);

	msg_add_via(msg, via);

	return(OK);
}


static ret_t contact_dup(void *value, u32_t n, void *arg1, void *arg2)
{
	address_t *contact  = (address_t *)value;
	msg_t *msg = (msg_t *)arg1;

	if(!contact || !msg)
		return(ERR);

	msg_add_contact(msg, contact);

	return(OK);
}


/**
 * Frees all memory associated with a sip msg object.
 *
 * @param msg SIP message object to be freed.
 */
static void msg_free(msg_t *msg)
{
	if(!msg)
		return;

	msg_request_free(msg->request);
	msg_response_free(msg->response);
	sapi_list_free(msg->headers, msg_param_free);
	sapi_list_free(msg->vias, msg_via_free);
	msg_address_free(msg->to);
	msg_address_free(msg->from);
	sapi_list_free(msg->contacts, msg_address_free);
	msg_cseq_free(msg->cseq);
	sapi_free(msg->call_id);
	sapi_free(msg->max_forwards);
	sapi_free(msg->expires);
	sapi_free(msg->user_agent);
	sapi_free(msg->ssid);
	sapi_free(msg->body);
	sapi_free(msg);
}


/**
 * Allocate and initialize a new SIP msg object.
 *
 * @return msg A newly created SIP msg object
 */
msg_t *msg_alloc(void)
{
	msg_t *msg   = (msg_t *)sapi_malloc((u32_t)sizeof(msg_t));

	msg->nrefs   = 1;
	sapi_mutex_init(&msg->mutex);

	return(msg);
}




/**
 * Increment reference count for "automatic" garbage collection"
 *
 * @param msg SIP message object to be referenced.
 * @return msg SIP message object to be referenced.
 */
const msg_t *msg_ref(const msg_t *msg)
{
	if(!msg)
		return(NULL);

	sapi_mutex_lock(&((msg_t *)msg)->mutex);
	((msg_t *)msg)->nrefs++;
	sapi_mutex_unlock(&((msg_t *)msg)->mutex);

	return(msg);
}


/**
 * Decrement reference count for "automatic" garbage collection"
 * Free if reference count reach zero.
 *
 * @param msg SIP message object to be dereferenced.
 */
void msg_deref(const msg_t *msg)
{
	u32_t nrefs;

	if(!msg)
		return;

	sapi_mutex_lock(&((msg_t *)msg)->mutex);
	nrefs = --((msg_t *)msg)->nrefs;
	sapi_mutex_unlock(&((msg_t *)msg)->mutex);

	if(!nrefs)
		msg_free((msg_t *)msg);
}


void msg_vderef(void *ptr)
{
	msg_deref((msg_t *)ptr);
}


msg_t *msg_dup(const msg_t *msg)
{
	msg_t *new;

	if(!msg || (!msg->request && !msg->response))
		return(NULL);

	new = msg_alloc();

	if(msg->request)
		msg_set_request(new, msg->request->method, msg->request->url);
	else
		msg_set_response(new, msg->response->statuscode, msg->response->reasonphrase);
	
	/* Headers */
	if(msg->headers)
		new->headers = msg_param_list_dup(msg->headers);
	
	/* Vias */
	if(msg->vias)
		sapi_list_apply(msg->vias, via_dup, new, NULL);
	
	/* To */
	if(msg->to)
		msg_set_to(new, msg->to);

	/* From */
	if(msg->from)
		msg_set_from(new, msg->from);

	/* Contacts */
	if(msg->contacts)
		sapi_list_apply(msg->contacts, contact_dup, new, NULL);

	/* Call-ID */
	if(msg->call_id)
		msg_set_call_id(new, msg->call_id);

	/* CSeq */
	if(msg->cseq)
		msg_set_cseq(new, msg->cseq->num, msg->cseq->method);
	
	/* Max-Forwards */
	if(msg->max_forwards)
		msg_set_max_forwards(new, *(msg->max_forwards));

	/* Expires */
	if(msg->expires)
		msg_set_expires(new, *(msg->expires));
	
	/* User-Agent */
	if(msg->user_agent)
		msg_set_user_agent(new, msg->user_agent);

	/* SSID */
	if(msg->ssid)
		msg_set_ssid(new, msg->ssid);

	/* Body */
	if(msg->body)
		msg_set_body(new, msg->body);

	return(new);
}


void msg_set_request(msg_t *msg, const u8_t *method, const url_t *url)
{
	if(!msg || !method || !url)
		return;
	
	if(msg->request)
		msg_request_free(msg->request);

	msg->request = msg_request_alloc();

	msg->request->method = sapi_strdup(method);
	msg->request->url    = msg_url_dup(url);
	msg->request->ver    = sapi_strdup(SIP_VERSION);
}


void msg_set_response(msg_t *msg, const u8_t *statuscode, const u8_t *reasonphrase)
{
	if(!msg || !statuscode || !reasonphrase)
		return;
	
	if(msg->response)
		msg_response_free(msg->response);

	msg->response = msg_response_alloc();

	msg->response->ver          = sapi_strdup(SIP_VERSION);
	msg->response->statuscode   = sapi_strdup(statuscode);
	msg->response->reasonphrase = sapi_strdup(reasonphrase);
}


void msg_add_via(msg_t *msg, const via_t *via)
{
	if(!msg || !via)
		return;
	
	if(!msg->vias)
		msg->vias = sapi_list_alloc();

	sapi_list_append(msg->vias, (void *)msg_via_dup(via));
}


void msg_prepend_via(msg_t *msg, const via_t *via)
{
	if(!msg || !via)
		return;
	
	if(!msg->vias)
		msg->vias = sapi_list_alloc();

	sapi_list_prepend(msg->vias, (void *)msg_via_dup(via));
}


void msg_set_to(msg_t *msg, const address_t *to)
{
	if(!msg || !to)
		return;

	if(msg->to)
		msg_address_free(msg->to);

	msg->to = msg_address_dup(to);
}


void msg_set_from(msg_t *msg, const address_t *from)
{
	if(!msg || !from)
		return;

	if(msg->from)
		msg_address_free(msg->from);

	msg->from = msg_address_dup(from);
}


void msg_add_contact(msg_t *msg, const address_t *contact)
{
	if(!msg || !contact)
		return;

	if(!msg->contacts)
		msg->contacts = sapi_list_alloc();
	
	sapi_list_append(msg->contacts, (void *)msg_address_dup(contact));
}


void msg_set_call_id(msg_t *msg, const u8_t *call_id)
{
	if(!msg || !call_id)
		return;

	if(msg->call_id)
		sapi_free(msg->call_id);

	msg->call_id = sapi_strdup(call_id);
}


void msg_set_cseq(msg_t *msg, u32_t num, const u8_t *method)
{
	if(!msg || !method)
		return;
	
	if(msg->cseq)
		msg_cseq_free(msg->cseq);

	msg->cseq = msg_cseq_alloc();

	msg->cseq->num    = num;
	msg->cseq->method = sapi_strdup(method);
}


void msg_set_max_forwards(msg_t *msg, u32_t max_forwards)
{
	if(!msg)
		return;

	if(msg->max_forwards)
		sapi_free(msg->max_forwards);

	msg->max_forwards = (u32_t *)sapi_malloc((u32_t)sizeof(u32_t));

	*(msg->max_forwards) = max_forwards;
}


void msg_dec_max_forwards(msg_t *msg)
{
	if(!msg || !msg->max_forwards)
		return;

	(*(msg->max_forwards))--;
}


void msg_set_expires(msg_t *msg, u32_t expires)
{
	if(!msg)
		return;

	if(msg->expires)
		sapi_free(msg->expires);

	msg->expires = (u32_t *)sapi_malloc((u32_t)sizeof(u32_t));

	*(msg->expires) = expires;
}


void msg_set_user_agent(msg_t *msg, const u8_t *user_agent)
{
	if(!msg || !user_agent)
		return;

	if(msg->user_agent)
		sapi_free(msg->user_agent);

	msg->user_agent = sapi_strdup(user_agent);
}


/**
 * Set the SSID on a SIP message.
 *
 * @param msg SIP message to be changed.
 * @param ssid SSID to be set.
 */
void msg_set_ssid(msg_t *msg, const u8_t *ssid)
{
	if(!msg || !ssid)
		return;

	if(msg->ssid)
		sapi_free(msg->ssid);

	msg->ssid = sapi_strdup(ssid);
}


void msg_add_header(msg_t *msg, const u8_t *hname, const u8_t *hvalue)
{
	if(!msg || !hname || !hvalue)
		return;

	if(!msg->headers)
		msg->headers = sapi_list_alloc();
	
	msg_param_list_add_param(msg->headers, hname, hvalue);
}


void msg_add_header_u32(msg_t *msg, const u8_t *hname, u32_t hvalue)
{
	if(!msg || !hname)
		return;

	if(!msg->headers)
		msg->headers = sapi_list_alloc();
	
	msg_param_list_add_param_u32(msg->headers, hname, hvalue);
}


void msg_set_body(msg_t *msg, const u8_t *body)
{
	if(!msg || !body) {
		MSG_DEBUG_PRINTF("msg_body_set(): msg or body is NULL");
		return;
	}
	
	if(msg->body)
		sapi_free(msg->body);

	msg->body = sapi_strdup(body);
}


via_t *msg_get_topvia(const msg_t *msg)
{
	if(!msg || !msg->vias)
		return(NULL);

	return((via_t *)sapi_list_get_head(msg->vias));
}


void msg_del_topvia(msg_t *msg)
{
	if(!msg || !msg->vias)
		return;

	sapi_list_del_head(msg->vias, msg_via_free);
}


u32_t msg_nvias(const msg_t *msg)
{
	if(!msg || !msg->vias)
		return(0);

	return(sapi_list_nnodes(msg->vias));
}


msg_t *msg_response(const msg_t *req, const u8_t *statuscode, const u8_t *reasonphrase)
{
	msg_t *resp;

	if(!req || !statuscode || !reasonphrase) {
		MSG_DEBUG_PRINTF("msg_response(): req, statuscode or reasonphrase is NULL");
		return(NULL);
	}

	resp = msg_alloc();

	msg_set_response(resp, statuscode, reasonphrase);

	sapi_list_apply(req->vias, via_dup, resp, NULL);

	msg_set_to(resp, req->to);
	msg_set_from(resp, req->from);
	msg_set_call_id(resp, req->call_id);
	msg_set_cseq(resp, req->cseq->num, req->cseq->method);

	return(resp);
}


msg_t *msg_cancel(const msg_t *req)
{
	msg_t *cancel;

	if(!req) {
		MSG_DEBUG_PRINTF("msg_cancel(): req is NULL");
		return(NULL);
	}

	cancel = msg_alloc();

	msg_set_request(cancel, "CANCEL", req->request->url);
	msg_add_via(cancel, msg_get_topvia(req));
	msg_set_to(cancel, req->to);
	msg_set_from(cancel, req->from);
	msg_set_call_id(cancel, req->call_id);
	msg_set_cseq(cancel, req->cseq->num, "CANCEL");

	return(cancel);
}


msg_t *msg_ack(const msg_t *req, const msg_t *resp)
{
	msg_t *ack;

	if(!req || !req->request || !resp || !resp->response) {
		MSG_DEBUG_PRINTF("msg_ack(): req, req->request, resp or resp->response is NULL");
		return(NULL);
	}

	ack = msg_alloc();

	msg_set_request(ack, "ACK", req->request->url);
	msg_add_via(ack, (via_t *)req->vias->start->value);
	msg_set_to(ack, resp->to);
	msg_set_from(ack, req->from);
	msg_set_call_id(ack, req->call_id);
	msg_set_cseq(ack, req->cseq->num, "ACK");

	return(ack);
}


void msg_dump(const msg_t *msg)
{
	u8_t *str = msg_str(msg);

	sapi_printf(str);
	sapi_free(str);
}
