#include <sapi.h>
#include <msg.h>


static ret_t startline_parse(msg_t *msg, const u8_t *msg_str, u8_t **headers_start)
{
	u8_t *cr, *lf, *startline_str;
	ret_t ret;

	cr = sapi_strchr(msg_str, '\r');
	lf = sapi_strchr(msg_str, '\n');

	if(lf - cr != 1) {
		MSG_DEBUG_PRINTF("startline_parse(): CRLF error");
		return(ERR);
	}

	(*headers_start) = lf + 1;
	startline_str = sapi_strndup(msg_str, (u32_t)(cr - msg_str));
	
	if(sapi_strncmp(startline_str, "SIP/", sapi_strlen("SIP/")) == FALSE) {
		msg->request = msg_request_alloc();
		ret = msg_request_parse(msg->request, startline_str);
	}
	else {
		msg->response = msg_response_alloc();
		ret = msg_response_parse(msg->response, startline_str);
	}

	sapi_free(startline_str);
	return(ret);
}


static ret_t headers_parse(list_t *headers, u8_t *headers_start, u8_t **body_start)
{
	param_t *header;
	u8_t *start, *vstart, *colon, *cr, *lf;

	start = headers_start;

	while(*start) {

		colon = sapi_strchr(start, ':');
		cr    = sapi_strchr(start, '\r');
		lf    = sapi_strchr(start, '\n');

		/* End of header ? */
		if(!colon || colon > cr) {
			if(start == cr && lf == cr + 1) {
				(*body_start) = *(lf+1) ? (lf+1) : NULL; /* have body ? */
				return(OK);
			}
			else {
				MSG_DEBUG_PRINTF("header_list_parse(): end of message linefeed error");
				break;
			}
		}

		/* Integrity check */
		if((colon == start) || (cr < colon + 1) || (cr + 1 != lf)) {
			MSG_DEBUG_PRINTF("header_list_parse(): integrity check failed");
			break;
		}

		header = msg_param_alloc();
		header->name  = sapi_strndup(start, colon - start);

		/* Remove leading spaces in header */
		vstart = colon + 1;
		while(*vstart == ' ')
			vstart++;

		header->value = sapi_strndup(vstart, cr - vstart);
		header->node = sapi_list_append(headers, header);

		start = lf + 1;
	}

	return(ERR);
}


static ret_t headers_apply(void *value, u32_t n, void *arg1, void *arg2)
{
	param_t *header = (param_t *)value;
	msg_t *msg = (msg_t *)arg1;

	if(!header || !msg || !header->name || !header->value || !header->node) {
		MSG_DEBUG_PRINTF("header_list_apply(): header or msg is NULL");
		return(ERR);
	}

	/* Via(s): */
	if(sapi_strcmp_ci(header->name, "Via") ||
	   sapi_strcmp_ci(header->name, "v")) {
		
		via_t *via;
		
		if(!msg->vias)
			msg->vias = sapi_list_alloc();
		
		via = msg_via_alloc();
		
		if(msg_via_parse(via, header->value) != OK) {
			msg_via_free(via);
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing Via header");
			return(ERR);
		}
		sapi_list_append(msg->vias, via);
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* To: */
	else if(!msg->to && (sapi_strcmp_ci(header->name, "To") ||
			     sapi_strcmp_ci(header->name, "t"))
			) {
		
		msg->to = msg_address_alloc();
		
		if(msg_address_parse(msg->to, header->value) != OK) {
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing To header");
			return(ERR);
		}
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* From: */
	else if(!msg->from && (sapi_strcmp_ci(header->name, "From") ||
			       sapi_strcmp_ci(header->name, "f"))) {
		
		msg->from = msg_address_alloc();
		
		if(msg_address_parse(msg->from, header->value) != OK) {
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing From header");
			return(ERR);
		}
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* Contact(s): */
	else if(sapi_strcmp_ci(header->name, "Contact") ||
		sapi_strcmp_ci(header->name, "m")) {
		
		address_t *contact;
		
		if(!msg->contacts)
			msg->contacts = sapi_list_alloc();
		
		contact = msg_address_alloc();
		
		/* Special case for 'Contact: *' */
		if(sapi_strcmp(header->value, "*") == TRUE) {
			contact->displayname = sapi_strdup("*");
		}
		else if(msg_address_parse(contact, header->value) != OK) {
			msg_address_free(contact);
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing Contact header");
			return(ERR);
		}
		sapi_list_append(msg->contacts, contact);
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* CSeq: */
	else if(!msg->cseq && (sapi_strcmp_ci(header->name, "CSeq"))) {
		
		msg->cseq = msg_cseq_alloc();
		
		if(msg_cseq_parse(msg->cseq, header->value) != OK) {
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing CSeq header");
			return(ERR);
		}
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* Call-ID: */
	else if(!msg->call_id && (sapi_strcmp_ci(header->name, "Call-ID") ||
				  sapi_strcmp_ci(header->name, "i"))) {
		
		msg->call_id = sapi_strdup(header->value);
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* Max-Forwards: */
	else if(!msg->max_forwards && (sapi_strcmp_ci(header->name, "Max-Forwards"))) {
		
		msg->max_forwards = (u32_t *)sapi_malloc(sizeof(u32_t));
		
		if(sapi_to_number3(msg->max_forwards, header->value, sapi_strlen(header->value)) != OK) {
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing Max-Forwards header");
			return(ERR);
		}
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* Expires: */
	else if(!msg->expires && sapi_strcmp_ci(header->name, "Expires")) {
		
		msg->expires = (u32_t *)sapi_malloc(sizeof(u32_t));

		if(sapi_to_number3(msg->expires, header->value, sapi_strlen(header->value)) != OK) {
			MSG_DEBUG_PRINTF("header_list_apply(): error parsing Expires header");
			return(ERR);
		}
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* User-Agent */
	else if(!msg->user_agent && sapi_strcmp_ci(header->name, "User-Agent")) {
		
		msg->user_agent = sapi_strdup(header->value);
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}
	/* SSID */
	else if(!msg->ssid && sapi_strcmp_ci(header->name, "X-SSID")) {
		
		msg->ssid = sapi_strdup(header->value);
		sapi_list_remove(msg->headers, header->node, msg_param_free);
	}


	return(OK);
}


/**
 * Parse a SIP message payload string into a SIP object.
 *
 * @param msg     SIP message object to be populated.
 * @param msg_str String containing the SIP message.
 * @return OK if successfully parsed, ERR otherwise.
 */
ret_t msg_parse(msg_t *msg, const u8_t *msg_str)
{
	u8_t *headers_start, *body_start;
	
	/* Parse startline */
	if(startline_parse(msg, msg_str, &headers_start) == ERR){
		MSG_DEBUG_PRINTF("msg_parse(): could not parse startline");
		return(ERR);
	}

	msg->headers = sapi_list_alloc();

	/* Parse headers into a name: value list */
	if(headers_parse(msg->headers, headers_start, &body_start) == ERR){
		MSG_DEBUG_PRINTF("msg_parse(): could not parse header list");
		return(ERR);
	}

	/* Parse known headers in detail, unknown headers are kept in msg->headers */
	if(sapi_list_apply(msg->headers, headers_apply, msg, NULL) != OK) {
		MSG_DEBUG_PRINTF("msg_parse(): Error parsing headers");
		return(ERR);
	}

	/* Save body */
	if(body_start)
		msg->body = sapi_strdup(body_start);
	
	/* Integrity check */
	if(!msg->to || !msg->from || !msg->vias || !msg->cseq || !msg->call_id) {
		MSG_DEBUG_PRINTF("msg_parse(): Mandatory header fields is missing");
		return(ERR);
	}

	return(OK);
}
