#include <sapi.h>
#include <msg.h>


static ret_t via_apply(void *value, u32_t n, void *arg1, void *arg2)
{
	via_t *via = (via_t *)value;
	u8_t **str = (u8_t **)arg1, *tmp;

	if(!via || !str) {
		MSG_DEBUG_PRINTF("via_apply(): via or str is NULL");
		return(ERR);
	}

	if((tmp = msg_via_str(via))) {
		sapi_saprintf(str, "Via: %s\r\n", tmp);
		sapi_free(tmp);
	}
	else {
		MSG_DEBUG_PRINTF("via_apply(): msg_via_str(): failed");
		return(ERR);
	}

	return(OK);
}


static ret_t contact_apply(void *value, u32_t n, void *arg1, void *arg2)
{
	address_t *contact = (address_t *)value;
	u8_t **str = (u8_t **)arg1, *tmp;

	if(!contact || !str) {
		MSG_DEBUG_PRINTF("contact_apply(): contact or str is NULL");
		return(ERR);
	}

	/* Special case for 'Contact: *' */
	if(sapi_strcmp(contact->displayname, "*")) {
		sapi_saprintf(str, "Contact: *\r\n");
	}
	else if((tmp = msg_address_str(contact))) {
		sapi_saprintf(str, "Contact: %s\r\n", tmp);
		sapi_free(tmp);
	}
	else {
		MSG_DEBUG_PRINTF("contact_apply(): msg_via_str(): failed");
		return(ERR);
	}

	return(OK);
}


static ret_t header_apply(void *value, u32_t n, void *arg1, void *arg2)
{
	param_t *header = (param_t *)value;
	u8_t **str = (u8_t **)arg1;

	if(!header || !str) {
		MSG_DEBUG_PRINTF("unparsed_header_apply(): header or str is NULL");
		return(ERR);
	}

	sapi_saprintf(str, "%s: %s\r\n", header->name, header->value);
	
	return(OK);
}


/**
 * Allocates and creates a string containg the SIP message object.
 * Handles both requests and responses.
 *
 * @param msg SIP message object to read from.
 * @return String containing the SIP message.
 */
u8_t *msg_str(const msg_t *msg)
{
	u8_t *str = NULL, *tmp;

	if(!msg) {
		MSG_DEBUG_PRINTF("msg_str(): msg is NULL");
		goto out_err;
	}

	/* Integrity check */
	if(!msg->vias || !msg->to || !msg->from || !msg->cseq || !msg->call_id) {
		MSG_DEBUG_PRINTF("msg_str(): Mandatory header fields is missing");
		goto out_err;
	}


	/* Request */
	if(msg->request) {
		if((tmp = msg_request_str(msg->request))) {
			sapi_saprintf(&str, "%s\r\n", tmp);
			sapi_free(tmp);
		}
		else {
			MSG_DEBUG_PRINTF("msg_str(): msg_request_str(): failed");
			goto out_err;
		}
	}
	/* Response */
	else if(msg->response) {
		if((tmp = msg_response_str(msg->response))) {
			sapi_saprintf(&str, "%s\r\n", tmp);
			sapi_free(tmp);
		}
		else {
			MSG_DEBUG_PRINTF("msg_str(): msg_response_str(): failed");
			goto out_err;
		}
	}
	else {
		MSG_DEBUG_PRINTF("msg_str(): Msg is neither response or request");
		goto out_err;
	}

	/* Via(s): */
	if(sapi_list_apply(msg->vias, via_apply, &str, NULL) != OK)
		goto out_err;

	/* To: */
	if((tmp = msg_address_str(msg->to))) {
		sapi_saprintf(&str, "To: %s\r\n", tmp);
		sapi_free(tmp);
	}
	else {
		MSG_DEBUG_PRINTF("msg_str(): msg_address_str(): failed for To header");
		goto out_err;		
	}

	/* From: */
	if((tmp = msg_address_str(msg->from))) {
		sapi_saprintf(&str, "From: %s\r\n", tmp);
		sapi_free(tmp);
	}
	else {
		MSG_DEBUG_PRINTF("msg_str(): msg_address_str(): failed for From header");
		goto out_err;		
	}

	/* Contact(s): */
	if(msg->contacts) {
		if(sapi_list_apply(msg->contacts, contact_apply, &str, NULL) != OK)
			goto out_err;
	}

	/* Call-ID: */
	sapi_saprintf(&str, "Call-ID: %s\r\n", msg->call_id);

	/* CSeq: */
	if((tmp = msg_cseq_str(msg->cseq))) {
		sapi_saprintf(&str, "CSeq: %s\r\n", tmp);
		sapi_free(tmp);
	}
	else {
		MSG_DEBUG_PRINTF("msg_str(): msg_cseq_str(): failed");
		goto out_err;
	}

	/* Max-Forwards */
	if(msg->max_forwards) {
		sapi_saprintf(&str, "Max-Forwards: %i\r\n", *(msg->max_forwards));
	}

	/* Expires */
	if(msg->expires) {
		sapi_saprintf(&str, "Expires: %i\r\n", *(msg->expires));
	}	

	/* User-Agent: */
	if(msg->user_agent) {
		sapi_saprintf(&str, "User-Agent: %s\r\n", msg->user_agent);
	}

	/* SSID: */
	if(msg->ssid) {
		sapi_saprintf(&str, "X-SSID: %s\r\n", msg->ssid);
	}

	/* Unparsed Headers */
	if(msg->headers) {
		if(sapi_list_apply(msg->headers, header_apply, &str, NULL) != OK)
			goto out_err;
	}

	/* End Of Headers */
	sapi_saprintf(&str, "\r\n");

	/* Body */
	if(msg->body)
		sapi_saprintf(&str, "%s", msg->body);


	return(str);

out_err:
	sapi_free(str);
	return(NULL);
}
