#include <sapi.h>
#include <msg.h>


url_t *msg_url_alloc()
{
	return((url_t *)sapi_malloc((u32_t)sizeof(url_t)));
}


void msg_url_free(url_t *url)
{
	if(!url)
		return;

	sapi_free(url->scheme);
	sapi_free(url->username);
	sapi_free(url->password);
	sapi_free(url->host);
	sapi_list_free(url->params, msg_param_free);
	sapi_list_free(url->headers, msg_param_free);

	sapi_free(url);
}


url_t *msg_url_dup(const url_t *url)
{
	url_t *new_url;

	if(!url)
		return(NULL);

	new_url = msg_url_alloc();
	
	if(url->scheme)
		new_url->scheme = sapi_strdup(url->scheme);
	
	if(url->username)
		new_url->username = sapi_strdup(url->username);

	if(url->password)
		new_url->password = sapi_strdup(url->password);

	if(url->host)
		new_url->host = sapi_strdup(url->host);

	new_url->port = url->port;

	if(url->params)
		new_url->params = msg_param_list_dup(url->params);

	if(url->headers)
		new_url->headers = msg_param_list_dup(url->headers);
	
	return(new_url);
}


ret_t msg_url_parse(url_t *url, const u8_t *url_str)
{
	u8_t *colon1, *colon2, *colon3, *at, *host_start, *host_end, *semicolon, *question;
	u8_t *url_end, *port_end;


	if(!url || !url_str)
		return(ERR);

	/* Scheme */
	colon1 = sapi_strchr(url_str, ':');
	if(!colon1) {
		MSG_DEBUG_PRINTF("url_parse(): colon1 missing");
		return(ERR);
	}

	url->scheme = sapi_strndup(url_str, colon1 - url_str);

	/* Username / Password */
	at = sapi_strchr(colon1, '@');
	if(at) {
		colon2 = sapi_strchr(colon1 + 1, ':');
		if(colon2 && colon2 < at) {
			url->username = sapi_lower(sapi_strndup(colon1 + 1, colon2 - colon1 - 1));
			url->password = sapi_lower(sapi_strndup(colon2 + 1, at - colon2 - 1));
		}
		else {
			url->username = sapi_lower(sapi_strndup(colon1 + 1, at - colon1 - 1));
		}
		host_start = at + 1;
	}
	else {
		host_start = colon1 + 1;
	}


	/* Host */
	if(!*host_start) {
		MSG_DEBUG_PRINTF("url_parse(): premature end");
		return(ERR);
	}

	colon3    = sapi_strchr(host_start, ':');
	semicolon = sapi_strchr(host_start, ';');
	question  = sapi_strchr(host_start, '?');
	url_end   = (u8_t *)url_str + sapi_strlen(url_str);

	if(colon3)
		host_end = colon3;
	else if(semicolon && (!question || (semicolon < question)))
		host_end = semicolon;
	else if(question)
		host_end = question;
	else
		host_end = url_end;

	url->host = sapi_strndup(host_start, host_end - host_start);

	/* Port */
	if(colon3) {
		if(semicolon) {
			if(semicolon < colon3) {
				MSG_DEBUG_PRINTF("url_parse(): semicolon < colon3");
				return(ERR);
			}
			else
				port_end = semicolon;
		}
		else if(question) {
			if(question < colon3) {
				MSG_DEBUG_PRINTF("url_parse(): question < colon3");
				return(ERR);
			}
			else
				port_end = question;
		}
		else {
			port_end = url_end;
		}

		url->port = sapi_to_number(colon3 + 1, port_end - 1);
		if(!url->port) {
			MSG_DEBUG_PRINTF("url_parse(): invalid port");
			return(ERR);
		}
	}
	else {
		url->port = 5060;
	}

	/* Params */
	if(semicolon && (!question || (semicolon < question))) {
		if(question) {
			if(question < semicolon) {
				MSG_DEBUG_PRINTF("url_parse(): question < semicolon");
				return(ERR);
			}
			else {
				url->params = sapi_list_alloc();
				if(msg_param_list_parse(url->params, semicolon + 1, question, ';') != OK) {
					MSG_DEBUG_PRINTF("url_parse(): error parsing url params (1)");
					return(ERR);
				}
			}
		}
		else {
			url->params = sapi_list_alloc();
			if(msg_param_list_parse(url->params, semicolon + 1, url_end, ';') != OK) {
				MSG_DEBUG_PRINTF("url_parse(): error parsing url params (2)");
				return(ERR);
			}
		}
	}

	/* Headers */
	if(question) {
		url->headers = sapi_list_alloc();
		if(msg_param_list_parse(url->headers, question + 1, url_end, '&') != OK) {
			MSG_DEBUG_PRINTF("url_parse(): error parsing url headers");
			return(ERR);
		}
	}


	/* Integrity check */
	if(!url->host || !url->port) {
		MSG_DEBUG_PRINTF("url_parse(): Mandatory fields is missing");
		return(ERR);
	}


	return(OK);
}


u8_t *msg_url_str(const url_t *url)
{
	u8_t *str = NULL, *tmp;

	if(!url || !url->scheme || !url->host) {
		MSG_DEBUG_PRINTF("msg_url_str(): url, url->scheme or url->host is NULL");
		return(NULL);
	}

	sapi_saprintf(&str, "%s:", url->scheme);

	if(url->username) {
		sapi_saprintf(&str, "%s", url->username);
		if(url->password)
			sapi_saprintf(&str, ":%s", url->password);
		
		sapi_saprintf(&str, "@");
	}

	sapi_saprintf(&str, "%s", url->host);

	if(url->port != 5060)
		sapi_saprintf(&str, ":%u", url->port);

	if(url->params) {
		tmp = msg_param_list_str(url->params, ';');
		sapi_saprintf(&str, ";%s", tmp);
		sapi_free(tmp);
	}

	if(url->headers) {
		tmp = msg_param_list_str(url->headers, '&');
		sapi_saprintf(&str, "?%s", tmp);
		sapi_free(tmp);
	}	
	
	return(str);
}


u8_t *msg_url_cannonical(const url_t *url)
{
	u8_t *cannon = NULL;

	if(!url || !url->scheme || !url->host) {
		MSG_DEBUG_PRINTF("msg_url_cannonical(): url, url->scheme or url->host is NULL");
		return(NULL);
	}

	sapi_saprintf(&cannon, "%s:", url->scheme);

	if(url->username) {
		sapi_saprintf(&cannon, "%s", url->username);
		if(url->password)
			sapi_saprintf(&cannon, ":%s", url->password);
		
		sapi_saprintf(&cannon, "@");
	}

	sapi_saprintf(&cannon, "%s", url->host);

	if(url->port != 5060)
		sapi_saprintf(&cannon, ":%u", url->port);

	return(cannon);
}
