#include <sapi.h>
#include <msg.h>


via_t *msg_via_alloc()
{
	return((via_t *)sapi_malloc((u32_t)sizeof(via_t)));
}


void msg_via_free(void *ptr)
{
	via_t *via = (via_t *)ptr;

	if(!via)
		return;

	sapi_free(via->ver);
	sapi_free(via->proto);
	sapi_free(via->host);
	sapi_list_free(via->params, msg_param_free);
	sapi_free(via);
}


via_t *msg_via_dup(const via_t *via)
{
	via_t *new_via;

	if(!via)
		return(NULL);

	new_via = msg_via_alloc();

	if(via->ver)
		new_via->ver = sapi_strdup(via->ver);

	if(via->proto)
		new_via->proto = sapi_strdup(via->proto);
		
	if(via->host)
		new_via->host = sapi_strdup(via->host);

	new_via->port = via->port;

	if(via->params)
		new_via->params = msg_param_list_dup(via->params);

	return(new_via);
}


via_t *msg_via_create(const u8_t *host, u16_t port)
{
	via_t *via = msg_via_alloc();

	via->ver    = sapi_strdup(SIP_VERSION);
	via->proto  = sapi_strdup("UDP");
	via->host   = sapi_strdup(host);
	via->port   = port;

	return(via);
}


void msg_via_addparam(via_t *via, const u8_t *name, const u8_t *value)
{
	if(!via || !name || !value) {
		MSG_DEBUG_PRINTF("msg_via_addparam(): via, name of value is NULL");
		return;
	}

	if(!via->params)
		via->params = sapi_list_alloc();

	msg_param_list_add_param(via->params, name, value);
}


u8_t *msg_via_getparam(const via_t *via, const u8_t *name)
{
	if(!via || !name) {
		MSG_DEBUG_PRINTF("msg_via_getparam(): via or name is NULL");
		return(NULL);
	}

	return(msg_param_list_get_param(via->params, name));
}


ret_t msg_via_parse(via_t *via, const u8_t *via_str)
{
	u8_t *slash1, *slash2, *space, *colon, *semicolon, *host_end, *port_end, *end;

	if(!via || !via_str)
		return(ERR);

	slash1 = sapi_strchr(via_str, '/');
	if(!slash1) {
		MSG_DEBUG_PRINTF("msg_via_parse(): First slash is missing");
		return(ERR);
	}

	slash2 = sapi_strchr(slash1 + 1, '/');
	if(!slash2) {
		MSG_DEBUG_PRINTF("msg_via_parse(): Second slash is missing");
		return(ERR);
	}

	space = sapi_strchr(slash2, ' ');
	if(!space) {
		MSG_DEBUG_PRINTF("msg_via_parse(): Space is missing");
		return(ERR);
	}
	
	via->ver   = sapi_strndup(via_str, slash2 - via_str);

	if(sapi_strcmp(via->ver, SIP_VERSION) == FALSE) {
		MSG_DEBUG_PRINTF("msg_via_parse(): Unsupported SIP Version: '%s'", via->ver);
		return(ERR);
	}

	via->proto = sapi_strndup(slash2 + 1, space - slash2 - 1);

	colon     = sapi_strchr(space, ':');
	semicolon = sapi_strchr(space, ';');
	end       = (u8_t *)via_str + sapi_strlen(via_str);

	/* Host */
	if(colon) {
		host_end = colon;
	}
	else if(semicolon)
		host_end = semicolon;
	else
		host_end = end;

	via->host = sapi_strndup(space + 1, host_end - space - 1);

	/* Port */
	if(colon) {
		if(semicolon)
			port_end = semicolon;
		else
			port_end = end;
		
		if(sapi_to_number3((u32_t *)&(via->port), colon + 1, port_end - colon - 1) != OK) {
			MSG_DEBUG_PRINTF("msg_via_parse(): Error parsing port");
			return(ERR);
		}
	}
	else {
		via->port = SIP_PORT;
	}

	/* Params */
	if(semicolon) {
		via->params = sapi_list_alloc();
		if(msg_param_list_parse(via->params, semicolon + 1, end, ';') != OK)
			return(ERR);
	}

	/* Integrity check */
	if(!via->ver || !via->proto || !via->host || !via->port) {
		MSG_DEBUG_PRINTF("via_parse(): Mandatory fields is missing");
		return(ERR);
	}

	return(OK);
}


u8_t *msg_via_str(const via_t *via)
{
	u8_t *str = NULL, *tmp;

	if(!via || !via->ver || !via->proto || !via->host) {
		MSG_DEBUG_PRINTF("msg_via_str(): via or via->ver or via->proto or via->host is NULL");
		return(NULL);
	}

	sapi_saprintf(&str, "%s/%s %s:%i", via->ver, via->proto, via->host, via->port);

	if(via->params) {
		tmp = msg_param_list_str(via->params, ';');
		sapi_saprintf(&str, ";%s", tmp);
		sapi_free(tmp);
	}

	return(str);
}
