#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>
#include <sapi.h>
#include <rtp.h>


typedef struct {
	u8_t *host;
	u16_t port;
	void *payload;
	size_t payload_size;	
} packet_t;


/** The version of RTP we support */
#define RTP_VERSION 2


/**
 * RTP header for little-endian machines.
 */
struct rtp_header {
	u8_t  csrc_count:4;   /**< The CSRC count contains the number of CSRC
				 identifiers that follow the fixed header. */
	u8_t  extension:1;    /**< If the extension bit is set, the fixed header
				 is followed by one header extension. */
	u8_t  padding:1;      /**< 1 bit of padding. */
	u8_t  version:2;      /**< 2 bits of RTP version number (normally has
				 value 2). */


	u8_t  payload_type:7; /**< Identifies the RTP payload type. */
	u8_t  marker:1;       /**< The marker is intended to allow significant
				 events to be marked in the packet. */

	u16_t seq;            /**< Sequence number. */
	u32_t timestamp;      /**< The timestamp reflects the sampling instant
				 of the first octet in the RTP packet. */
	u32_t ssrc;           /**< The SSRC field identifies the synchronization
				 source. */
};


static rtp_recv_t __recv  = NULL;
static void *__recv_param = NULL;
static fifo_t *__sendq    = NULL;
static u16_t __rtp_port = 0;
pthread_t __receiver;
pthread_t __sender;


static inline packet_t *packet_alloc()
{
	return((packet_t *)sapi_malloc(sizeof(packet_t)));
}


static inline void packet_free(packet_t *packet)
{
	if(!packet)
		return;
	
	sapi_free(packet->host);
	sapi_free(packet->payload);
	sapi_free(packet);
}


static void *rtp_recv_thread(void *arg)
{
	int fd = -1, size, on = 1;
	struct sockaddr_in sa, ca;
	socklen_t ca_len = sizeof(ca);
	char buf[64000];

	if(__rtp_port == 0)
		goto out;

	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		sapi_logf("rtp_recv_thread(): socket(): %s", strerror(errno));
		goto out;
	}

	bzero(&(sa), sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons((unsigned short int)__rtp_port);

	if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
		sapi_logf("rtp_recv_thread(): setsockopt(): %s", strerror(errno));
		goto out;
	}

	if(bind(fd, (struct sockaddr *)&sa, (socklen_t)sizeof(sa)) < 0) {
		sapi_logf("rtp_recv_thread(): bind(): %s", strerror(errno));
		goto out;
	}
	
	for(;;) {
		size = recvfrom(fd, buf, sizeof(buf) - 1, 0, (struct sockaddr *)&ca, &ca_len);

		if(__recv)
			__recv((void *)buf, size, __recv_param);
	}

 out:
	RTP_DEBUG_PRINTF("rtp_recv_thread(): exit");

	if(fd != -1)
		close(fd);

	return(NULL);
}


static void *rtp_send_thread(void *arg)
{
	int fd = -1;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(fd == -1) {
                RTP_DEBUG_PRINTF("rtp_send_thread(): socket(): %s", strerror(errno));
		goto out;
        }
	
	while(__sendq) {

		packet_t *p = (packet_t *)sapi_fifo_get(__sendq);
		struct hostent hent, *hres;
		u8_t hbuf[8192];
		s32_t h_errno;
		u32_t addr;
		struct sockaddr_in sa;
		
		if(gethostbyname_r(p->host, &hent, hbuf, sizeof(hbuf), &hres, &h_errno)) {

			RTP_DEBUG_PRINTF("rtp_send_thread(): %s: %s", hstrerror(h_errno), p->host);
			goto err;
		}

		bcopy(hres->h_addr_list[0], (char *)&addr, sizeof(addr));
		endhostent();

		bzero((void *)&sa, sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_port   = htons((unsigned short int)p->port);
		sa.sin_addr.s_addr = addr;

		if(sendto(fd, (void *)p->payload, p->payload_size,
			  0, (struct sockaddr *)&sa, (socklen_t)sizeof(sa)) < 0) {

			RTP_DEBUG_PRINTF("rtp_send_thread(): %s", strerror(errno));
			goto err;
		}
		
		// RTP_DEBUG_PRINTF("rtp_send_thread(): packet sent to %s:%u", p->host, p->port);
		
	err:
		packet_free(p);
	}

 out:
	if(fd != -1)
		close(fd);

	RTP_DEBUG_PRINTF("rtp_send_thread(): exit");
	
	return(NULL);
}


void rtp_init(u16_t port, rtp_recv_t recv, void *param)
{
	__recv       = recv;
	__recv_param = param;
	__sendq      = sapi_fifo_alloc();
	__rtp_port  = port;

	pthread_create(&__receiver, NULL, rtp_recv_thread, NULL);
	pthread_create(&__sender, NULL, rtp_send_thread, NULL);

	RTP_DEBUG_PRINTF("RTP service is up and running");
}


static void rtp_header_populate(u8_t *dst, struct rtp_header *rtph)
{
	memcpy(dst, rtph, sizeof(struct rtp_header));
}


void rtp_send(const u8_t *host, u16_t port, const void *payload, size_t payload_size, u8_t pt)
{
	packet_t *p;
	struct rtp_header rtph;
	static int seq = 0, ts = 0;

	if(!__sendq) {
		RTP_DEBUG_PRINTF("rtp_send(): send queue not initialized");
		return;
	}
	
	if(!host || !port || !payload || !payload_size) {
		RTP_DEBUG_PRINTF("rtp_send(): host, port or payload is NULL");
		return;
	}

	p = packet_alloc();

	p->host    = sapi_strdup(host);
	p->port    = port;

	p->payload = sapi_malloc(sizeof(struct rtp_header) + payload_size);

	/* header */
	rtph.version      = RTP_VERSION;
	rtph.padding      = 0;
	rtph.extension    = 0;
	rtph.csrc_count   = 0;
	rtph.marker       = 0;
	rtph.payload_type = pt;
	rtph.seq          = htons(seq++);
	rtph.timestamp    = htonl(ts);
	rtph.ssrc         = 0xdeadbeef;

	ts += 160;

	rtp_header_populate(p->payload, &rtph);
	memcpy(p->payload + sizeof(struct rtp_header), payload, payload_size);

	p->payload_size = sizeof(struct rtp_header) + payload_size;

	sapi_fifo_put(__sendq, p);
}


void rtp_status(void)
{
	if(!__sendq)
		return;

	sapi_printf("RTP Send Queue:");
	sapi_fifo_status(__sendq);
}
