#include <string.h>
#include <pthread.h>
#include "sapi.h"


fifo_t *sapi_fifo_alloc(void)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
        pthread_cond_t cond  = PTHREAD_COND_INITIALIZER;
	fifo_t *fifo = (fifo_t *)sapi_malloc(sizeof(fifo_t));

        memcpy(&fifo->mutex, &mutex, sizeof(pthread_mutex_t));
        memcpy(&fifo->cond, &cond, sizeof(pthread_cond_t));

	return(fifo);
}


void sapi_fifo_free(fifo_t *fifo, fifo_free_t freefunc)
{
	node_t *node;

	if(!fifo) {
		SAPI_DEBUG_PRINTF("sapi_fifo_free(): fifo is NULL");
		return;
	}

	node = fifo->tail;

	while(node) {
		if(freefunc)
			freefunc(node->value);

		sapi_free(node);
		node = node->next;
	}

	sapi_free(fifo);
}


void sapi_fifo_put(fifo_t *fifo, const void *value)
{
	node_t *node;

	if(!fifo || !value) {
		SAPI_DEBUG_PRINTF("sapi_fifo_put(): fifo or value is NULL");
		return;
	}

	node = (node_t *)sapi_malloc(sizeof(node_t));

	node->next  = NULL;
	node->value = (void *)value;

	pthread_mutex_lock(&fifo->mutex);

	if(fifo->head)
		fifo->head->next = node;

	fifo->head = node;

	if(!fifo->tail)
		fifo->tail = node;

	fifo->n++;
	fifo->tot++;
	if(fifo->n > fifo->max)
		fifo->max = fifo->n;

	pthread_mutex_unlock(&fifo->mutex);
	pthread_cond_signal(&fifo->cond);
}


void *sapi_fifo_get(fifo_t *fifo)
{
	void *value = NULL;
	node_t *node;

	if(!fifo) {
		SAPI_DEBUG_PRINTF("sapi_fifo_get(): fifo is NULL");
		return(NULL);
	}

	pthread_mutex_lock(&fifo->mutex);

	while(!fifo->tail)
		pthread_cond_wait(&fifo->cond, &fifo->mutex);

	node = fifo->tail;

	fifo->tail = fifo->tail->next;
	if(!fifo->tail)
		fifo->head = NULL;
		
	value = node->value;
	sapi_free(node);
	fifo->n--;
	
	pthread_mutex_unlock(&fifo->mutex);

	return(value);
}


u32_t sapi_fifo_nnodes(fifo_t *fifo)
{
	u32_t nnodes;

	if(!fifo) {
		SAPI_DEBUG_PRINTF("sapi_fifo_status(): fifo is NULL");
		return(0);
	}

	pthread_mutex_lock(&fifo->mutex);
        nnodes = fifo->n;
	pthread_mutex_unlock(&fifo->mutex);

	return(nnodes);
}


void sapi_fifo_status(fifo_t *fifo)
{
	if(!fifo) {
		SAPI_DEBUG_PRINTF("sapi_fifo_status(): fifo is NULL");
		return;
	}

	pthread_mutex_lock(&fifo->mutex);
        sapi_printf(" FIFO Status:");
        sapi_printf("  Current: %8i", fifo->n);
        sapi_printf("  Total:   %8i", fifo->tot);
        sapi_printf("  Max:     %8i", fifo->max);
	pthread_mutex_unlock(&fifo->mutex);
}


u8_t *sapi_fifo_status_str(fifo_t *fifo)
{
	u8_t *str = NULL;

	if(!fifo) {
		return(sapi_sprintf("sapi_fifo_status(): fifo is NULL"));
	}

	pthread_mutex_lock(&fifo->mutex);

	sapi_saprintf(&str, "    <tr bgcolor=\"1478EB\">\n");
	sapi_saprintf(&str, "     <td colspan=\"2\"><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>FIFO Status</b></font></td>\n");
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Current</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", fifo->n);
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Total</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", fifo->tot);
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Max</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", fifo->max);
	sapi_saprintf(&str, "    </tr>\n");

	pthread_mutex_unlock(&fifo->mutex);

	return(str);
}
