#include "sapi.h"


#define HASH_PRIME (5381)


typedef struct {
	u8_t *key;
	void *value;
	node_t *node;
	hash_free_t free;
} entry_t;


static inline entry_t *entry_alloc(void)
{
	return((entry_t *)sapi_malloc(sizeof(entry_t)));
}


static inline void entry_free(void *ptr)
{
	entry_t *entry = (entry_t *)ptr;

	if(!entry) {
		SAPI_DEBUG_PRINTF("entry_free(): entry is NULL");
		return;
	}

	if(entry->free)
		entry->free(entry->value);

	sapi_free(entry->key);
	sapi_free(entry);
}


static bool_t entry_find(void *value, void *arg)
{
	entry_t *entry = (entry_t *)value;
	u8_t *key = (u8_t *)arg;

	if(!entry || !key) {
		SAPI_DEBUG_PRINTF("entry_find(): entry or key is NULL");
		return(FALSE);
	}

	return(sapi_strcmp(entry->key, key));
}


static inline u32_t hash(const u8_t *key)
{
	u32_t h = HASH_PRIME;

	while(*key) {
		h += (h<<5);
		h ^= (u32_t)*key++;
	}

	return(h);
}


hash_t *sapi_hash_alloc(u32_t size)
{
	hash_t *ht = sapi_malloc(sizeof(hash_t));
	u32_t rr = 3;

	while ((1 << rr) < size)
		rr++;

	ht->size    = 1 << rr;
	ht->buckets = sapi_malloc(ht->size * sizeof(list_t *));

	return(ht);
}


void sapi_hash_free(hash_t *ht)
{
	u32_t i;

	if(!ht) {
		SAPI_DEBUG_PRINTF("sapi_hash_free(): ht is NULL");
		return;
	}

	if(ht->buckets) {
		for(i=0; i<ht->size; i++) {
			if(ht->buckets[i]) {
				sapi_list_free(ht->buckets[i], entry_free);
			}
		}
		sapi_free(ht->buckets);
	}
	sapi_free(ht);
}


ret_t sapi_hash_ins(hash_t *ht, const u8_t *key, void *value, hash_free_t free)
{
	u32_t index;
	entry_t *entry;
	u32_t bucket_n;
	ret_t ret = ERR;

	if(!ht || !key || !value) {
		SAPI_DEBUG_PRINTF("sapi_hash_ins(): ht, key or value is NULL");
		return(ERR);
	}	

	index = hash(key) & (ht->size - 1);

	if(ht->buckets[index]) {
		if(sapi_list_find(ht->buckets[index], entry_find, (void *)key)) {
			SAPI_DEBUG_PRINTF("sapi_hash_ins(): Tried to insert duplicate key value");
			goto out;
		}
	}
	else
		ht->buckets[index] = sapi_list_alloc();
	
	entry        = entry_alloc();
	entry->key   = sapi_strdup(key);
	entry->value = value;
	entry->free  = free;
	entry->node  = sapi_list_append(ht->buckets[index], entry);
	
	/* Statistics */
	ht->n++;
	ht->tot++;
	if(ht->n > ht->max)
		ht->max = ht->n;

	bucket_n = sapi_list_nnodes(ht->buckets[index]);

	if(bucket_n > ht->bucket_max)
		ht->bucket_max = bucket_n;

	ret = OK;
out:
	return(ret);
}


void *sapi_hash_find(const hash_t *ht, const u8_t *key)
{
	u32_t index;
	entry_t *entry;
	void *value = NULL;

	if(!ht || !key) {
		SAPI_DEBUG_PRINTF("sapi_hash_get(): ht or key is NULL");
		return(NULL);
	}	

	index = hash(key) & (ht->size - 1);

	if(!ht->buckets[index])
		goto out;
	
	entry = (entry_t *)sapi_list_find(ht->buckets[index], entry_find, (void *)key);
	if(!entry)
		goto out;

	value = entry->value;

out:
	return(value);
}


void sapi_hash_del(hash_t *ht, const u8_t *key)
{
	u32_t index;
	entry_t *entry;

	if(!ht || !key) {
		SAPI_DEBUG_PRINTF("sapi_hash_del(): ht or key is NULL");
		return;
	}	

	index = hash(key) & (ht->size - 1);

	if(!ht->buckets[index]) {
		SAPI_DEBUG_PRINTF("sapi_hash_del(): No such entry '%s'", key);
		return;
	}

	entry = (entry_t *)sapi_list_find(ht->buckets[index], entry_find, (void *)key);
	if(!entry) {
		SAPI_DEBUG_PRINTF("sapi_hash_del(): No such entry '%s'", key);
		return;
	}

	sapi_list_remove(ht->buckets[index], entry->node, entry_free);

	/* This code is for debugging memory leaks; to be removed */
#if 1
	if(!sapi_list_nnodes(ht->buckets[index])) {
		sapi_list_free(ht->buckets[index], NULL);
		ht->buckets[index] = NULL;
	}
#endif

	/* Statistics */
	ht->n--;

}


static ret_t hash_apply(void *value, u32_t n, void *arg1, void *arg2)
{
	entry_t *entry = (entry_t *)value;
	hash_apply_t apply = (hash_apply_t)arg2;

	if(!entry || !apply) {
		SAPI_DEBUG_PRINTF("hash_apply(): entry or apply is NULL");
		return(ERR);
	}	

	return(apply(entry->value, arg1));
}


ret_t sapi_hash_apply(hash_t *ht, hash_apply_t apply, void *arg)
{
	u32_t i;
	ret_t ret = ERR;

	if(!ht || !apply) {
		SAPI_DEBUG_PRINTF("sapi_hash_apply(): ht or apply is NULL");
		return(ERR);
	}

	for(i=0; i<ht->size; i++) {
		if(ht->buckets[i]) {
			if(sapi_list_apply(ht->buckets[i], hash_apply, arg, apply) != OK)
				goto out;
		}
	}

	ret = OK;
out:
	return(ret);
}


void sapi_hash_status(const hash_t *ht)
{
	if(!ht) {
		SAPI_DEBUG_PRINTF("sapi_hash_status(): ht is NULL");
		return;
	}

        sapi_printf(" Hash Table Status:");
	sapi_printf("  Buckets:    %8i", ht->size);
        sapi_printf("  Current:    %8i", ht->n);
        sapi_printf("  Total:      %8i", ht->tot);
        sapi_printf("  Max:        %8i", ht->max);
	sapi_printf("  Bucket Max: %8i", ht->bucket_max);
}


u8_t *sapi_hash_status_str(const hash_t *ht)
{
	u8_t *str = NULL;

	if(!ht) {
		return(sapi_sprintf("sapi_hash_status(): ht is NULL"));
	}

	sapi_saprintf(&str, "    <tr bgcolor=\"1478EB\">\n");
	sapi_saprintf(&str, "     <td colspan=\"2\"><font face=\"Arial,Helvetica\" color=\"#ffffff\" size=\"-1\"><b>Hash Table Status</b></font></td>\n");
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Buckets</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", ht->size);
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Current</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", ht->n);
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Total</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", ht->tot);
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Max</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", ht->max);
	sapi_saprintf(&str, "    </tr>\n");

	sapi_saprintf(&str, "    <tr bgcolor=\"#ffffff\">\n");
	sapi_saprintf(&str, "     <td nowrap><font face=\"Arial,Helvetica\" size=\"-1\">Bucket Max</font></td>\n");
	sapi_saprintf(&str, "     <td align=\"right\" nowrap><font face=\"Arial,Helvetica\" size=\"-1\">%i</font></td>\n", ht->bucket_max);
	sapi_saprintf(&str, "    </tr>\n");

	return(str);
}



static ret_t entry_dump(void *value, u32_t n, void *arg1, void *arg2)
{
	entry_t *entry = (entry_t *)value;

	if(!entry) {
		SAPI_DEBUG_PRINTF("entry_dump(): entry is NULL");
		return(ERR);
	}

	sapi_printf("  Entry '%s'", entry->key);

	return(OK);
}


void sapi_hash_dump(const hash_t *ht)
{
	u32_t i;

	if(!ht) {
		SAPI_DEBUG_PRINTF("sapi_hash_dump(): ht is NULL");
		return;
	}

	sapi_printf("Hash Size: %i", ht->size);

	for(i=0; i<ht->size; i++) {
		if(ht->buckets[i]) {
			sapi_printf(" Bucket %i", i);
			sapi_list_apply(ht->buckets[i], entry_dump, NULL, NULL);
		}
	}
}


void sapi_hash_histogram(const hash_t *ht)
{
	u32_t i;

	if(!ht) {
		SAPI_DEBUG_PRINTF("sapi_hash_histogram(): ht is NULL");
		return;
	}

	sapi_printf(" Hash Histogram: (%i buckets)", ht->size);

	for(i=0; i<ht->size; i++) { 

		sapi_printf2("  Bucket %03i: ", i);

		if(ht->buckets[i]) {
			u32_t j;

			for(j=0; j < sapi_list_nnodes(ht->buckets[i]); j++)
				sapi_printf2("#");

			sapi_printf2(" (%i)", sapi_list_nnodes(ht->buckets[i]));
		}
		
		sapi_printf2("\n");
	}
}
