#include <stdarg.h>
#include "sapi.h"


static node_t *node_alloc(void)
{
	return((node_t *)sapi_malloc((u32_t)sizeof(node_t)));
}


static void node_free(node_t *node, free_t free)
{
	if(!node) {
		SAPI_DEBUG_PRINTF("node_free(): node is NULL");
		return;
	}

	if(free)
		free(node->value);

	sapi_free(node);
}


list_t *sapi_list_alloc(void)
{
	return((list_t *)sapi_malloc((u32_t)sizeof(list_t)));
}


void sapi_list_free(list_t *list, free_t free)
{
	node_t *curr, *tmp;

	if(!list) {
		// SAPI_DEBUG_PRINTF("sapi_list_free(): list is NULL");
		return;
	}

	curr = list->start;

	while(curr) {
		tmp = curr;
		curr = curr->next;
		node_free(tmp, free);
	}
	sapi_free(list);
}


node_t *sapi_list_prepend(list_t *list, void *value)
{
	node_t *new_node;
	
	if(!list || !value) {
		SAPI_DEBUG_PRINTF("sapi_list_appendx(): list or value is NULL");
		return(NULL);
	}

	new_node = node_alloc();

	new_node->value = value;

	/* List is empty */
	if(!list->start) {
		list->start     = new_node;
		list->end       = new_node;
	}
	/* List contains at least one node */
	else {
		list->start->prev = new_node;
		new_node->next    = list->start;
		list->start       = new_node;
	}

	list->n++;
	list->tot++;
	if(list->n > list->max)
		list->max = list->n;

	return(new_node);
}


node_t *sapi_list_append(list_t *list, void *value)
{
	node_t *new_node;

	if(!list || !value) {
		SAPI_DEBUG_PRINTF("sapi_list_appendx(): list or value is NULL");
		return(NULL);
	}

	new_node = node_alloc();

	new_node->value = value;

	/* List is empty */
	if(!list->start) {
		list->start     = new_node;
		list->end       = new_node;
	}
	/* List contains at least one node */
	else {
		list->end->next = new_node;
		new_node->prev  = list->end;
		list->end       = new_node;
	}

	list->n++;
	list->tot++;
	if(list->n > list->max)
		list->max = list->n;

	return(new_node);
}


void sapi_list_remove(list_t *list, node_t *node, free_t free)
{
	if(!list || !node) {
		SAPI_DEBUG_PRINTF("sapi_list_removex(): list or node is NULL");
		return;
	}

	/* node is first node */
	if(node == list->start)
		list->start = list->start->next;

	/* node is last node */
	if(node == list->end)
		list->end = list->end->prev;

	/* node before */
	if(node->prev)
		node->prev->next = node->next;

	/* node after */
	if(node->next)
		node->next->prev = node->prev;

	node_free(node, free);
	list->n--;
}


ret_t sapi_list_apply(list_t *list, list_apply_t apply, void *arg1, void *arg2)
{
	node_t *node;
	u32_t n = 0;

	if(!list || !apply) {
		SAPI_DEBUG_PRINTF("sapi_list_apply(): list or apply is NULL");
		return(OK);
	}

	node = list->start;

	while(node) {
		void *value = node->value;
		node = node->next;
		if(apply(value, n++, arg1, arg2) != OK)
			return(ERR);
	}

	return(OK);
}


void *sapi_list_find(list_t *list, list_find_t find, void *arg)
{
	node_t *node;

	if(!list || !find) {
		SAPI_DEBUG_PRINTF("sapi_list_find(): list or find is NULL");
		return(NULL);
	}

	node = list->start;

	while(node) {
		void *value = node->value;
		node = node->next;
		if(find(value, arg))
			return(value);
	}

	return(NULL);
}


void *sapi_list_get_head(const list_t *list)
{
	if(!list || !list->start)
		return(NULL);

	return(list->start->value);
}


void sapi_list_del_head(list_t *list, free_t free)
{
	if(!list || !list->start)
		return;

	sapi_list_remove(list, list->start, free);
}


u32_t sapi_list_nnodes(const list_t *list)
{
	if(!list) {
		SAPI_DEBUG_PRINTF("sapi_list_nnodes(): list is NULL");
		return(0);
	}

	return(list->n);
}


void sapi_list_status(const list_t *list)
{
	if(!list) {
		SAPI_DEBUG_PRINTF("sapi_list_status(): list is NULL");
		return;
	}

        sapi_printf(" LIST Status:");
        sapi_printf("  Current: %8i", list->n);
        sapi_printf("  Total:   %8i", list->tot);
        sapi_printf("  Max:     %8i", list->max);
}


u8_t *sapi_list_status_str(const list_t *list)
{
	u8_t *str = NULL;

	if(!list) {
		return(sapi_sprintf("sapi_list_status(): list is NULL"));
	}

        sapi_saprintf(&str, " LIST Status:\n");
        sapi_saprintf(&str, "  Current: %8i\n", list->n);
        sapi_saprintf(&str, "  Total:   %8i\n", list->tot);
        sapi_saprintf(&str, "  Max:     %8i\n", list->max);

	return(str);
}


void sapi_list_dump(const list_t *list)
{
	node_t *node;

	if(!list) {
		SAPI_DEBUG_PRINTF("sapi_list_dump(): list is NULL");
		return;
	}

	sapi_printf("nnodes: %i", list->n);
	node = list->start;

	while(node) {
		sapi_printf(" node @ 0x%08X", (u32_t)node);
		sapi_printf("  value: 0x%08X", (u32_t)node->value);
		sapi_printf("  prev : 0x%08X", (u32_t)node->prev);
		sapi_printf("  next : 0x%08X", (u32_t)node->next);

		node = node->next;
	}
}
