#include <pthread.h>
#include <unistd.h>
#include "sapi.h"


typedef struct {
	u32_t left;
	timeout_t timeout;
	void *arg;
	u8_t *sarg;
	node_t *node;
} oneshot_t;


static pthread_mutex_t __mutex = PTHREAD_MUTEX_INITIALIZER;
static list_t *__oneshotq = NULL;
static fifo_t *__attachq = NULL;
static fifo_t *__fireq = NULL;
static pthread_t __oneshot;
static pthread_t __attach;
static pthread_t __fire;


static inline oneshot_t *oneshot_alloc(void)
{
	return((oneshot_t *)sapi_malloc((sizeof(oneshot_t))));
}


static void oneshot_free(void *arg)
{
	oneshot_t *os = (oneshot_t *)arg;

	if(!os) {
		SAPI_DEBUG_PRINTF("oneshot_free(): os is NULL");
		return;
	}

	sapi_free(os->sarg);
	sapi_free(os);

	// SAPI_DEBUG_PRINTF("oneshot_free():");
}


static ret_t oneshot(void *value, u32_t n, void *arg1, void *arg2)
{
	oneshot_t *os = (oneshot_t *)value;

	if(!os) {
		SAPI_DEBUG_PRINTF("oneshot(): os is NULL");
		return(ERR);
	}

	if(!os->left) {
		
		sapi_list_remove(__oneshotq, os->node, NULL);

		sapi_fifo_put(__fireq, os);
	}
	else
		os->left--;

	return(OK);
}


static void *fire_thread(void *arg)
{
	while(__fireq) {
		
		oneshot_t *os = sapi_fifo_get(__fireq);
		
		if(os->timeout)
			os->timeout(os->arg, os->sarg);
		
		oneshot_free(os);
	}
	
	return(NULL);
}


static void *oneshot_thread(void *arg)
{
	struct timespec req;

	req.tv_sec  = 0;
	req.tv_nsec = 100000000 - 10000000;

	while(__oneshotq && __fireq) {

		pthread_mutex_lock(&__mutex);
		sapi_list_apply(__oneshotq, oneshot, NULL, NULL);
		pthread_mutex_unlock(&__mutex);

		nanosleep(&req, NULL);
	}
	
	return(NULL);
}


static void *attach_thread(void *arg)
{
	while(__attachq && __oneshotq) {
		
		oneshot_t *os = sapi_fifo_get(__attachq);
		
		SAPI_LOCK(&__mutex);
		os->node = sapi_list_append(__oneshotq, os);
		SAPI_UNLOCK(&__mutex);
	}
	
	return(NULL);
}


void sapi_oneshot_init(void)
{
	__attachq  = sapi_fifo_alloc();
	__oneshotq = sapi_list_alloc();
	__fireq    = sapi_fifo_alloc();
	pthread_create(&__oneshot, NULL, oneshot_thread, NULL);
	pthread_create(&__attach,  NULL, attach_thread,  NULL);
	pthread_create(&__fire,    NULL, fire_thread,  NULL);

	sapi_logf("Oneshot service is up and running");
}

void sapi_oneshot(u32_t time, timeout_t timeout, void *arg, const u8_t *sarg)
{
	oneshot_t *os;

	if(!__attachq) {
		SAPI_DEBUG_PRINTF("sapi_oneshot_attach():  __attachq is NULL");
		return;		
	}

	os = oneshot_alloc();

	os->left    = time;
	os->timeout = timeout;
	os->arg     = arg;
	os->sarg    = sapi_strdup(sarg);

	sapi_fifo_put(__attachq, os);
}


void sapi_oneshot_status(void)
{
	if(!__attachq || !__oneshotq || !__fireq) {
		SAPI_DEBUG_PRINTF("sapi_oneshot_status():  __attachq or __oneshotq is NULL");
		return;		
	}

	sapi_printf("Oneshot Status:");

	SAPI_LOCK(&__mutex);
	sapi_list_status(__oneshotq);
	SAPI_UNLOCK(&__mutex);

	sapi_fifo_status(__attachq);
	sapi_fifo_status(__fireq);
}


u8_t *sapi_oneshot_status_str(void)
{
	u8_t *str = NULL, *tmp;

	if(!__attachq || !__oneshotq || !__fireq) {
		return(sapi_sprintf("sapi_oneshot_status():  __attachq, __oneshotq or __fireq is NULL"));
	}

	SAPI_LOCK(&__mutex);
	tmp = sapi_list_status_str(__oneshotq);
	SAPI_UNLOCK(&__mutex);
	sapi_saprintf(&str, "%s", tmp);
	sapi_free(tmp);

	tmp = sapi_fifo_status_str(__attachq);
	sapi_saprintf(&str, "%s", tmp);
	sapi_free(tmp);

	tmp = sapi_fifo_status_str(__fireq);
	sapi_saprintf(&str, "%s", tmp);
	sapi_free(tmp);

	return(str);
}
