/**
 * @file media.c Implements SDP media handling.
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>.  All rights reserved.
 */
#include <sapi.h>
#include <sdp.h>
#include "media.h"


/*
 * Progmem strings
 */
static const u8_t attr_prefix[] = "a=";       /**< Attribute field prefix. */
static const u8_t attr_rtpmap[] = "rtpmap:";  /**< Attribute rtpmap string. */
static const u8_t attr_fmtp[]   = "fmtp:";    /**< Attribute fmtp string. */


/**
 * Free one attribute object.
 *
 * @param ptr Attribute to be freed.
 */
void attrib_free(void *ptr)
{
	sdp_attrib_t *a = (sdp_attrib_t *)ptr;

	sapi_free(a->rtpmap);
	sapi_free(a->fmtp);
	sapi_free(a);
}


/**
 * Allocate a new media object.
 *
 * @return A newly allocated media object.
 */
sdp_media_t *media_alloc(void)
{
	return(sapi_malloc(sizeof(sdp_media_t)));
}


/**
 * Free a media object.
 *
 * @param m Media object to be freed.
 */
void media_free(sdp_media_t *m)
{
	sapi_free(m->type);
	sapi_free(m->transp);
	sapi_list_free(m->attribs, attrib_free);
	sapi_free(m);
}


/**
 * Parse SDP media field.
 *
 * @param media  Media object to parse into.
 * @param body   SDP body string to parse.
 *
 * @return OK if successfully parsed, ERR otherwise.
 *
 * Example:
 *
 * <pre>
 * m=audio 1234 RTP/AVP 4 0
 * </pre>
 */
ret_t media_parse(sdp_media_t *media, const u8_t *body)
{
	const u8_t *sp1, *sp2, *sp3, *pos, *end;
	const u8_t *start = body;

	/* first, find all the spaces */
	sp1 = sapi_strchr(body, ' ');
	if(!sp1)
		goto error;

	sp2 = sapi_strchr(sp1 + 1, ' ');
	if(!sp2)
		goto error;

	sp3 = sapi_strchr(sp2 + 1, ' ');
	if(!sp3)
		goto error;

	end = sapi_strchr(start, '\r');
	if(!end)
		goto error;

	media->type = sapi_strndup(start, sp1 - start);

	media->port = (u16_t)sapi_to_number(sp1 + 1, sp2 - 1);

	media->transp = sapi_strndup(sp2 + 1, sp3 - sp2 - 1);

	pos = sp3;

	media->attribs = sapi_list_alloc();

	while(pos < end) {

		const u8_t *space = sapi_strchr(pos + 1, ' ');
		sdp_attrib_t *attrib = (sdp_attrib_t *)sapi_malloc(sizeof(sdp_attrib_t));

		if(space >= end) {
			space = end;
		}

		attrib->rtp_pt = sapi_to_number(pos + 1, space - 1);

		sapi_list_append(media->attribs, (void *)attrib);

		if(space == end)
			break;

		pos = space;
	}

	return(OK);

 error:
	SDP_DEBUGFx("parse error in media field");
	return(ERR);
}


/**
 * Parse SDP attribute field into a media object.
 *
 * @param media  Media object to parse into.
 * @param body   SDP body string to parse.
 *
 * @return OK if successfully parsed, ERR otherwise.
 *
 * Example:
 *
 * <pre>
 * m=audio 1234 RTP/AVP 4 0
 * a=rtpmap:4 G723/8000
 * a=fmtp:4 bitrate=6300
 * a=rtpmap:0 PCMU/8000
 * a=ptime:20
 * </pre>
 */
ret_t media_parse_attrib(sdp_media_t *media, const u8_t *body)
{
	const u8_t *colon, *space, *slash;
	const u8_t *start = body, *end;
	u8_t rtp_pt;
	node_t *node;

	ASSERT(media);

	colon = sapi_strchr(body, ':');

	if(!colon)
		goto error;

	space = sapi_strchr(colon, ' ');
	if(!space) {
		return(OK);
	}
	slash = sapi_strchr(space, '/');
	if(!slash) {
		return(OK);
	}
	end = sapi_strchr(space, '\n');
	if(!end) {
		return(OK);
	}

	rtp_pt = (u16_t)sapi_to_number(colon + 1, space - 1);

	/* find corresponding rtp_pt in list and append to list */
	node = media->attribs->start;

	while(node) {

		sdp_attrib_t * const attrib = (sdp_attrib_t *)node->value;

		if(rtp_pt == attrib->rtp_pt) {

			body = start;

			if(sapi_strcmp(body, attr_rtpmap) == OK) {

				if(!attrib->rtpmap) {
					attrib->rtpmap = sapi_strndup(space + 1, slash - space - 1);
					attrib->clockrate = sapi_to_number(slash + 1, end - 2);
				}
			}
			else if(sapi_strcmp(body, attr_fmtp) == OK) {

				if(!attrib->fmtp)
					attrib->fmtp = sapi_strndup(space + 1, end - space - 1);
			}
		}

		node = node->next;
	}

	return(OK);

 error:
	SDP_DEBUGFx("parse error in media attrib field");
	return(ERR);
}


/**
 * Print one attribute line to a netbuffer.
 *
 * @param body   The netbuffer to print to.
 * @param name   Name of attribute e.g. "rtpmap:".
 * @param rtp_pt RTP payload type.
 * @param value  Value of attribute e.g. "PCMU/8000".
 */
static void attr_str(u8_t *body, const u8_t *name, u8_t rtp_pt, const u8_t *value)
{
	(void)sapi_saprintf(&body, "%s", attr_prefix);
	(void)sapi_saprintf(&body, "%s", name);
	(void)sapi_saprintf(&body, "%u ", rtp_pt);
	(void)sapi_saprintf(&body, "%s", value);
}


/**
 * Print a attrib object to a string (list_apply callback).
 *
 * @param value Node value.
 * @param arg   Callback argument.
 */
static ret_t attrib_to_str(void *value, u32_t n, void *arg1, void *arg2)
{
	const sdp_attrib_t *attrib = (const sdp_attrib_t *)value;
	u8_t *body = (u8_t *)arg1;

	(void)n;
	(void)arg2;

	if(attrib->rtpmap) {
		attr_str(body, attr_rtpmap, attrib->rtp_pt, attrib->rtpmap);
		sapi_saprintf(&body, "/%u\r\n", attrib->clockrate);
	}

	if(attrib->fmtp) {
		attr_str(body, attr_fmtp, attrib->rtp_pt, attrib->fmtp);
		sapi_saprintf(&body, "\r\n");
	}

	return(OK);
}


/**
 * Print a media object to a netbuffer.
 *
 * @param media  The media object to be printed.
 * @param body   The netbuffer to print to.
 */
void media_str(sdp_media_t *media, u8_t *body)
{
	(void)sapi_list_apply(media->attribs, attrib_to_str, body, NULL);
}
