/**
 * @file sdp.c Implements SDP - Session Description Protocol (RFC 2327)
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>.  All rights reserved.
 */
#include <sapi.h>
#include <sdp.h>
#include "media.h"


/*
 * Program memory strings
 */
static const u8_t sdp_version[]      = "v=" SDP_VERSION "\r\n";  /**< Version field. */
static const u8_t sdp_owner[]        = "o=- 0 0 IN IP4 ";        /**< Owner field. */
static const u8_t sdp_session[]      = "s=-\r\n";                /**< Name field. */
static const u8_t sdp_conn[]         = "c=IN IP4 ";              /**< Connection field. */
static const u8_t sdp_time[]         = "t=0 0\r\n";              /**< Time field. */
static const u8_t sdp_media_prefix[] = "m=";                     /**< Media field prefix. */


/*
 * Private functions.
 */


/**
 * Parse SDP connection field.
 *
 * @param sdp    SDP object to parse into.
 * @param body   SDP body in netbuffer to parse.
 *
 * @return OK if successfully parsed, ERR otherwise.
 */
static inline ret_t sdp_parse_conn(sdp_t *sdp, const u8_t *body)
{
	const u8_t *sp1, *sp2, *end;

	sp1 = sapi_strchr(body, ' ');
	if(!sp1)
		goto error;

	sp2 = sapi_strchr(sp1 + 1, ' ');
	if(!sp2)
		goto error;

	end = sapi_strchr(sp1 + 1, '\r');
	if(!end)
		goto error;

	sdp->host = sapi_strndup(sp2 + 1, end - sp2 - 1);

	if(sdp->host)
		return(OK);

 error:
	SDP_DEBUGFx("parse error in connection field");
	return(ERR);
}


/*
 * Public functions.
 */


/**
 * Create a new SDP object.
 *
 * @param host          Connection host.
 * @param media_type    Media type e.g. "audio" or "video".
 * @param media_port    Media port.
 * @param media_transp  Media Transport type e.g. "RTP/AVP".
 *
 * @return  The newly created SDP object.
 */
sdp_t *sdp_create(const u8_t *host, const u8_t *media_type, u16_t media_port, const u8_t *media_transp)
{
	sdp_t *sdp;
	sdp_media_t *media;

	sdp = (sdp_t *)sapi_malloc(sizeof(sdp_t));

	sdp->host = sapi_strdup(host);

	media = media_alloc();
	media->type = sapi_strdup(media_type);
	media->port = media_port;
	media->transp = sapi_strdup(media_transp);

	media->attribs = sapi_list_alloc();

	sdp->media = media;

	return(sdp);
}


/**
 * Append an attribute to the SDP object.
 *
 * @param sdp           SDP object to append to.
 * @param rtp_pt        Attribute RTP payload type.
 * @param rtpmap        Attribute rtpmap string e.g. "PCMU".
 * @param clockrate     Clockrate e.g. 8000.
 * @param fmtp          Attribute format list with e.g. bit rates..
 */
void sdp_media_add(sdp_t *sdp, u8_t rtp_pt, const u8_t *rtpmap, u16_t clockrate, const u8_t *fmtp)
{
	sdp_attrib_t *attrib;

	ASSERT(sdp);

	attrib = (sdp_attrib_t *)sapi_malloc(sizeof(sdp_attrib_t));

	attrib->rtp_pt = rtp_pt;
	if(rtpmap) {
		attrib->rtpmap = sapi_strdup(rtpmap);
		attrib->clockrate = clockrate;
	}
	if(fmtp)
		attrib->fmtp   = sapi_strdup(fmtp);

	sapi_list_append(sdp->media->attribs, (void *)attrib);
}


/**
 * Write media attributes to netbuffer (list_apply callback)
 *
 * @param value Node value.
 * @param arg   Callback argument.
 */
static ret_t media_attrib_str(void *value, u32_t n, void *arg1, void *arg2)
{
	const sdp_attrib_t *attrib = (const sdp_attrib_t *)value;
	u8_t **body = (u8_t **)arg1;

	(void)n;
	(void)arg2;

	(void)sapi_saprintf(body, " %u", attrib->rtp_pt);

	return(OK);
}


/**
 * Print a SDP object to a netbuffer.
 *
 * @param sdp  The SDP object to print.
 *
 * @return  A new string with the SDP body printed
 *
 * Example:
 *
 * <pre>
 * v=0
 * o=- 0 0 IN IP4 10.47.10.66
 * s=-
 * c=IN IP4 10.47.10.66
 * t=0 0
 * m=audio 1234 RTP/AVP 4 0
 * a=rtpmap:0 PCMU/8000
 * a=ptime:20
 * </pre>
 */
u8_t *sdp_str(const sdp_t *sdp)
{
	u8_t *start = NULL;
	u8_t **body = &start;

	ASSERT(sdp);

	/* v= (protocol version) */
	(void)sapi_saprintf(body, "%s", sdp_version);

	/* o=  (owner/creator and session identifier). */
	(void)sapi_saprintf(body, "%s", sdp_owner);
	(void)sapi_saprintf(body, "%s\r\n", sdp->host);

	/* s=  (session name) */
	(void)sapi_saprintf(body, "%s", sdp_session);

	/* c=  (connection information - not required if included in all media) */
	(void)sapi_saprintf(body, "%s", sdp_conn);
	(void)sapi_saprintf(body, "%s\r\n", sdp->host);

	/* t=  (time the session is active) */
	(void)sapi_saprintf(body, "%s", sdp_time);

	/* m=  (media name and transport address) */
	(void)sapi_saprintf(body, "%s", sdp_media_prefix);
	(void)sapi_saprintf(body, "%s", sdp->media->type);
	(void)sapi_saprintf(body, " %u ", sdp->media->port);
	(void)sapi_saprintf(body, "%s", sdp->media->transp);

	/* media attribs to netbuf */
	(void)sapi_list_apply(sdp->media->attribs, media_attrib_str, body, NULL);

	(void)sapi_saprintf(body, "\r\n");

	/* a= (attribute field) */
	media_str(sdp->media, *body);

	return(start);
}


/**
 * Parse SDP body and return a new SDP object containing all values.
 *
 * @note The client must remember to call sdp_free() to free the SDP object.
 *
 * @param body String containing the SDP body to parse.
 *
 * @return A new SDP object if successfully parsed, NULL otherwise.
 */
sdp_t *sdp_parse(const u8_t *body)
{
	const u8_t *start, *equal, *cr, *lf;
	u8_t ch;
	sdp_t *sdp;
	ret_t ret = OK;

	if(!body)
		return(NULL);

	SDP_DEBUGFx("sdp_parse");

	sdp = (sdp_t *)sapi_malloc(sizeof(sdp_t));

	start = body;

	while(*start) {

		equal = sapi_strchr(start, '=');
		cr    = sapi_strchr(start, '\r');
		lf    = sapi_strchr(start, '\n');

		/* finished? */
		if(!equal) {
			return(sdp);
		}

		if(equal != start + 1 || lf != cr + 1) {
			goto error;
		}

		ch = *start;

		switch(ch){

		case 'c':
			if(!sdp->host)
				ret = sdp_parse_conn(sdp, start + 1);
			break;

		case 'm':
			if(sdp->media)
				break;
			else
				sdp->media = media_alloc();

			ret = media_parse(sdp->media, equal + 1);
			break;

		case 'a':
			ret = media_parse_attrib(sdp->media, equal + 1);
			break;

		} /* switch */

		if(ret != OK)
			goto error;

		start = lf + 1;
	}

	return(sdp);

 error:
	SDP_DEBUGFx("parse error - dumping");
	sdp_dump(sdp);
	sdp_free(sdp);
	return(NULL);
}


/**
 * Free an SDP object.
 *
 * @param sdp SDP object to be freed.
 */
void sdp_free(sdp_t *sdp)
{
	sapi_free(sdp->host);
	media_free(sdp->media);
	sapi_free(sdp);
}
