#include <stdlib.h> // for rand()
#include <time.h>   // for time()
#include <sapi.h>
#include <msg.h>
#include <tp.h>
#include <ctrans.h>
#include <strans.h>
#include <ldb.h>
#include "sipd.h"


#define PROXY_TIMEOUT (1800) // Response Context Timeout (180 seconds or 3 minutes)
#define PROXY_HASH_SIZE (32768)


typedef struct {
	tp_t *tp;
	msg_t *msg;
	bool_t got_final_resp;
} req_t;


typedef struct {
	const sipd_t *sipd;
	const msg_t *ireq;
	u8_t *hkey;
	list_t *reqs;
	bool_t got_final_resp;
} rctx_t;


static pthread_mutex_t __mutex = PTHREAD_MUTEX_INITIALIZER;
static hash_t *__ht = NULL;


static inline req_t *req_alloc()
{
	return((req_t *)sapi_malloc(sizeof(req_t)));
}


static void req_free(void *ptr)
{
	req_t *req = (req_t *)ptr;

	if(!req) {
		SIPD_DEBUG_PRINTF("req_free(): req is NULL");
		return;
	}

	tp_deref(req->tp);
	msg_deref(req->msg);
	
	sapi_free(req);
}


static inline rctx_t *rctx_alloc()
{
	return((rctx_t *)sapi_malloc(sizeof(rctx_t)));
}


static void rctx_free(void *ptr)
{
	rctx_t *rctx = (rctx_t *)ptr;

	if(!rctx) {
		SIPD_DEBUG_PRINTF("rctx_free(): rctx is NULL");
		return;
	}

	msg_deref(rctx->ireq);
	sapi_free(rctx->hkey);
	sapi_list_free(rctx->reqs, req_free);
	
	sapi_free(rctx);
}


static ret_t proxy_cancel(void *value, u32_t n, void *arg1, void *arg2)
{
	req_t *req = (req_t *)value;
	msg_t *cancel;
	
        if(!req) {
                SIPD_DEBUG_PRINTF("proxy_cancel(): req is NULL");
                return(ERR);
        }

	if(!req->got_final_resp) {
		cancel = msg_cancel(req->msg);
		ctrans_req(req->tp, cancel, NULL, NULL, NULL, NULL);
		msg_deref(cancel);			
		SIPD_DEBUG_PRINTF("proxy_cancel(): OK");
	}

        return(OK);
}


static void proxy_timeout(void *arg, u8_t *hkey)
{
	rctx_t *rctx;
	msg_t *resp;

	if(!hkey) {
		SIPD_DEBUG_PRINTF("proxy_timeout(): hkey is NULL");
		return;
	}

	SAPI_LOCK(&__mutex);

	rctx = sapi_hash_find(__ht, hkey);
	if(!rctx) {
		SIPD_DEBUG_PRINTF("proxy_timeout(): Response context not found");
		SAPI_UNLOCK(&__mutex);
		return;
	}

	if(sapi_strcmp(rctx->ireq->request->method, "INVITE"))
		sapi_list_apply(rctx->reqs, proxy_cancel, NULL, NULL);

	if(!rctx->got_final_resp) {
		resp = msg_response(rctx->ireq, "408", "Request Timeout");
		strans_resp(hkey, resp);
		msg_deref(resp);
	}

	sapi_hash_del(__ht, hkey);

	SAPI_UNLOCK(&__mutex);

	SIPD_DEBUG_PRINTF("proxy_timeout(): OK");
}


static void proxy_err(const msg_t *msg, const u8_t *reason, void *arg, u8_t *hkey)
{
        rctx_t *rctx;
        msg_t *oresp;

	if(!msg || !reason || !hkey) {
		SIPD_DEBUG_PRINTF("proxy_err(): msg, reason or hkey is NULL");
		return;
	}

	SAPI_LOCK(&__mutex);

	rctx = sapi_hash_find(__ht, hkey);
	if(!rctx) {
		SIPD_DEBUG_PRINTF("proxy_err(): Response context not found");
		SAPI_UNLOCK(&__mutex);
		return;
	}
	
	oresp = msg_response(rctx->ireq, "503", "Service Unavailable");
        strans_resp(hkey, oresp);
        msg_deref(oresp);

	SAPI_UNLOCK(&__mutex);

	SIPD_DEBUG_PRINTF("proxy_err(): %s", reason);
}


static void proxy_resp(const msg_t *msg, const msg_t *iresp, void *arg, u8_t *hkey)
{
        rctx_t *rctx;
        msg_t *oresp;
	req_t *req = (req_t *)arg;

	if(!msg || !iresp || !hkey || !req) {
		SIPD_DEBUG_PRINTF("proxy_resp(): msg, iresp, hkey or req is NULL");
		return;
	}

	/* Don't proxy 100 Responses */
        if(sapi_strcmp(iresp->response->statuscode, "100"))
                return;

	SAPI_LOCK(&__mutex);

	rctx = sapi_hash_find(__ht, hkey);
	if(!rctx) {
		SIPD_DEBUG_PRINTF("proxy_resp(): Response context for %s request not found", msg->request->method);
		SAPI_UNLOCK(&__mutex);
		return;
	}

	if(iresp->response->statuscode[0] != '1') {
		req->got_final_resp = TRUE;
		rctx->got_final_resp = TRUE;
	}

	/* Cancel all other INVITE requests if one of them is successful */
	if(sapi_strcmp(rctx->ireq->request->method, "INVITE") && sapi_strcmp(iresp->response->statuscode, "200"))
		sapi_list_apply(rctx->reqs, proxy_cancel, NULL, NULL);


	oresp = msg_dup(iresp);
        msg_del_topvia(oresp);
        strans_resp(hkey, oresp);
        msg_deref(oresp);

	SAPI_UNLOCK(&__mutex);

        SIPD_DEBUG_PRINTF("proxy_resp(): %s (%s)",
                          iresp->response->statuscode,
                          iresp->response->reasonphrase);
}


static ret_t proxy_fork(void *value, u32_t n, void *arg1, void *arg2)
{
	address_t *contact = (address_t *)value;
        rctx_t *rctx = (rctx_t *)arg1;
	req_t *req;
        via_t *via;
        u8_t *branch;

        if(!contact || !rctx) {
                SIPD_DEBUG_PRINTF("proxy_fork(): contact or rctx is NULL");
                return(ERR);
        }

	req = req_alloc();

        req->msg            = msg_dup(rctx->ireq);
	req->tp             = tp_alloc();
	req->got_final_resp = FALSE;
        
	req->tp->host  = sapi_strdup(contact->url->host);
	req->tp->port  = contact->url->port;
	req->tp->proto = UDP;

	msg_set_request(req->msg, rctx->ireq->request->method, contact->url);

        if(req->msg->max_forwards)
                msg_dec_max_forwards(req->msg);
	else
		msg_set_max_forwards(req->msg, SIPD_MAX_FORWARDS);

        via = msg_via_create(rctx->sipd->url->host, rctx->sipd->url->port);
        branch = sapi_sprintf("z9hG4bK%i%i", time(NULL), rand());
        msg_via_addparam(via, "branch", branch);
        sapi_free(branch);
        msg_prepend_via(req->msg, via);
        msg_via_free(via);
        
	sapi_list_append(rctx->reqs, req);

        ctrans_req(req->tp, req->msg, proxy_resp, proxy_err, req, rctx->hkey);
	
        SIPD_DEBUG_PRINTF("proxy_fork(): OK");

        return(OK);
}


void sipd_proxy_init(void)
{
	if(__ht)
		return;

	__ht = sapi_hash_alloc(PROXY_HASH_SIZE);
}


void sipd_proxy(const sipd_t *sipd, const msg_t *ireq, const u8_t *hkey)
{
	list_t *contacts = NULL;
	msg_t *resp;
	u8_t *addr;
	rctx_t *rctx;

	if(!__ht || !sipd || !ireq || !hkey) {
		SIPD_DEBUG_PRINTF("sipd_proxy(): __ht, sipd, ireq or hkey is NULL");
		return;
	}

	/* Check Max-Forwards */
	if(ireq->max_forwards && (*(ireq->max_forwards) == 0)) {
	
		/* 483 To Many Hops */
		resp = msg_response(ireq, "483", "To Many Hops");

		strans_resp(hkey, resp);
		msg_deref(resp);
		SIPD_DEBUG_PRINTF("sipd_proxy(): Max-Forwards is 0");
		return;
	}

	/* Local Request */
	if(sapi_strcmp_ci(ireq->to->url->host, sipd->domain)) {
		
		SIPD_DEBUG_PRINTF("sipd_proxy(): local request");

		addr     = msg_url_cannonical(ireq->to->url);
		contacts = ldb_get(addr);
		sapi_free(addr);
	}
	/* Remote Request */
	else {

		address_t *contact = msg_address_alloc();
		u8_t *str = msg_url_str(ireq->request->url);

		SIPD_DEBUG_PRINTF("sipd_proxy(): remote request");

		if(msg_address_parse(contact, str) == OK) {

			contacts = sapi_list_alloc();
			sapi_list_append(contacts, contact);
		}
		else {
			SIPD_DEBUG_PRINTF("sipd_proxy(): parsing request uri failed");
			msg_address_free(contact);
		}

		SIPD_DEBUG_PRINTF("sipd_proxy(): remote url: '%s'", str);
		sapi_free(str);
	}
	
	if(!contacts) {

		SIPD_DEBUG_PRINTF("sipd_proxy(): Not Found");

		resp = msg_response(ireq, "404", "Not Found");
		strans_resp(hkey, resp);
		msg_deref(resp);
	}
	else {
		SIPD_DEBUG_PRINTF("sipd_proxy(): Proxying %s request", ireq->request->method);

		/* Create New Response Context */
		rctx = rctx_alloc();
		rctx->sipd = (sipd_t *)sipd;
		rctx->ireq = msg_ref(ireq);
		rctx->hkey = sapi_strdup(hkey);
		rctx->reqs = sapi_list_alloc();
		
		SAPI_LOCK(&__mutex);

		if(sapi_hash_ins(__ht, hkey, rctx, rctx_free) != OK) {
			
			SIPD_DEBUG_PRINTF("sipd_proxy(): sapi_hash_ins(): failed");
			
			/* 500 Internal Server Error */
			resp = msg_response(ireq, "500", "Internal Server Error");
			strans_resp(hkey, resp);
			msg_deref(resp);
			
			rctx_free(rctx);
		}
		else {
			sapi_list_apply(contacts, proxy_fork, rctx, NULL);
			
			sapi_oneshot(PROXY_TIMEOUT, proxy_timeout, NULL, hkey);
		}

		SAPI_UNLOCK(&__mutex);
		sapi_list_free(contacts, msg_address_free);
	}
}


void sipd_proxy_cancel(const sipd_t *sipd, const msg_t *req, const msg_t *oreq, const u8_t *hkey, const u8_t *ohkey)
{
	msg_t *resp;

	if(!__ht || !sipd || !req || !hkey) {
		SIPD_DEBUG_PRINTF("sipd_proxy_cancel(): __ht, sipd, ireq or hkey is NULL");
		return;
	}

	if(ohkey && oreq) {

		rctx_t *rctx;

		SAPI_LOCK(&__mutex);

		rctx = sapi_hash_find(__ht, ohkey);
		if(!rctx) {
			SIPD_DEBUG_PRINTF("sipd_proxy_cancel(): Response context not found");
		}
		else {
			sapi_list_apply(rctx->reqs, proxy_cancel, NULL, NULL);
		}

		SAPI_UNLOCK(&__mutex);

		/* 200 to CANCEL Request */
		resp = msg_response(req, "200", "OK");
		strans_resp(hkey, resp);
		msg_deref(resp);
		
		/* 487 to INVITE Request */
		resp = msg_response(oreq, "487", "Request Terminated");			
		strans_resp(ohkey, resp);
		msg_deref(resp);

		SIPD_DEBUG_PRINTF("sipd_proxy_cancel(): Transaction canceled");
	}
	else {
		resp = msg_response(req, "481", "Transaction Does Not Exist");		
		strans_resp(hkey, resp);
		msg_deref(resp);

		SIPD_DEBUG_PRINTF("sipd_proxy_cancel(): Transaction Does Not Exist");
	}
}
