#include <stdlib.h> // for rand()
#include <time.h>   // for time()
#include <sapi.h>
#include <msg.h>
#include <tp.h>
#include <ctrans.h>
#include <strans.h>
#include <ldb.h>
#include "sipd.h"


static void tp_err(const u8_t *reason, u8_t *hkey)
{
	sapi_logf("Stateless Proxy Transport Error: %s", reason);
}


static ret_t stateless_tx(const sipd_t *sipd, const msg_t *ireq, const address_t *contact)
{
        via_t *via;
        u8_t *branch;
	msg_t *msg;
	tp_t *tp;

        if(!sipd || !ireq || !contact) {
                SIPD_DEBUG_PRINTF("stateless_tx(): contact or rctx is NULL");
                return(ERR);
        }

        msg = msg_dup(ireq);
	tp  = tp_alloc();
        
	tp->host  = sapi_strdup(contact->url->host);
	tp->port  = contact->url->port;
	tp->proto = UDP;

	msg_set_request(msg, ireq->request->method, contact->url);

        if(msg->max_forwards)
                msg_dec_max_forwards(msg);
	else
		msg_set_max_forwards(msg, SIPD_MAX_FORWARDS);

        via = msg_via_create(sipd->url->host, sipd->url->port);
        branch = sapi_sprintf("z9hG4bK%i%i", time(NULL), rand());
        msg_via_addparam(via, "branch", branch);
        sapi_free(branch);
        msg_prepend_via(msg, via);
        msg_via_free(via);
        
	tp_send(tp, msg, tp_err, NULL);

	tp_deref(tp);
	msg_deref(msg);

        SIPD_DEBUG_PRINTF("send_stateless(): OK");

        return(OK);
}


void sipd_proxy_sl_req(const sipd_t *sipd, const msg_t *ireq)
{
	address_t *contact = NULL;
	u8_t *addr;

	if(!sipd || !ireq) {
		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): __ht, sipd, ireq or hkey is NULL");
		return;
	}

	/* Check Max-Forwards */
	if(ireq->max_forwards && (*(ireq->max_forwards) == 0)) {
		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): Max-Forwards is 0");
		return;
	}

	/* Local Request */
	if(sapi_strcmp_ci(ireq->to->url->host, sipd->domain)) {

		list_t *contacts;		

		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): local request");

		addr     = msg_url_cannonical(ireq->to->url);
		contacts = ldb_get(addr);
		sapi_free(addr);

		if(contacts) {
			contact = msg_address_dup(sapi_list_get_head(contacts));
			sapi_list_free(contacts, msg_address_free);
		}
	}
	/* Remote Request */
	else {
		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): remote request");
		
		contact = msg_address_alloc();
		addr    = msg_url_str(ireq->request->url);

		if(msg_address_parse(contact, addr) != OK) {
			SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): parsing request uri failed");
			msg_address_free(contact);
			contact = NULL;
		}

		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): remote url: '%s'", addr);
		sapi_free(addr);
	}
	
	if(!contact) {
		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): Not Found");
		return;
	}
	else {
		SIPD_DEBUG_PRINTF("sipd_proxy_sl_req(): Proxying %s request", ireq->request->method);
		stateless_tx(sipd, ireq, contact);
		msg_address_free(contact);
	}
}


ret_t sipd_proxy_sl_resp(const sipd_t *sipd, const msg_t *iresp)
{
        msg_t *oresp = NULL;
	tp_t *tp = NULL;
	via_t *via;
	ret_t ret = ERR;

        if(!sipd || !iresp) {
                sapi_logf("sipd_proxy_sl_resp(): sipd or iresp is NULL");
		goto out;
        }

        oresp = msg_dup(iresp);
        msg_del_topvia(oresp);

	via = msg_get_topvia(oresp);
	if(!via) {
		sapi_logf("sipd_proxy_sl_resp(): No via header");
		goto out;	
	}

	tp  = tp_alloc();
        
	tp->host  = sapi_strdup(via->host);
	tp->port  = via->port;
	tp->proto = UDP;

	tp_send(tp, oresp, tp_err, NULL);

        SIPD_DEBUG_PRINTF("sipd_proxy_sl_resp(): OK");

	ret = OK;
out:
	msg_deref(oresp);
	tp_deref(tp);

        return(ret);
}
