#include <sapi.h>
#include <msg.h>
#include <tp.h>
#include <strans.h>
#include "st.h"


#define STRANS_HASH_SIZE (32768)
#define MAGIC_VIA_BRANCH "z9hG4bK"


#define T1   5 // 500 msecs
#define T2  40 //   4 secs
#define T4  50 //   5 secs


static hash_t *__ht = NULL;
static mutex_t __mutex = SAPI_MUTEX_INITIALIZER;


static inline u8_t *strans_hash_key_compat(const msg_t *req, bool_t cancel)
{
	u8_t *hkey = NULL, *tmp, *to_tag, *from_tag;
	via_t *via = msg_get_topvia(req);

	/* Request URI */
	tmp = msg_url_str(req->request->url);
	sapi_saprintf(&hkey, "%s", tmp);
	sapi_free(tmp);
       
	/* Call-ID and CSeq */
	if(cancel || sapi_strcmp(req->request->method, "ACK"))
		sapi_saprintf(&hkey, "|%s|%u|%s", req->call_id, req->cseq->num, "INVITE");
	else
		sapi_saprintf(&hkey, "|%s|%u|%s", req->call_id, req->cseq->num, req->request->method);
		
	/* Top VIA */
	tmp = msg_via_str(via);
	sapi_saprintf(&hkey, "|%s", tmp);
	sapi_free(tmp);

	/* To tag */
	to_tag = msg_address_getparam(req->to, "tag");
	if(to_tag) 
		sapi_saprintf(&hkey, "|%s", to_tag);
	
	/* From tag */
	from_tag = msg_address_getparam(req->from, "tag");
	if(from_tag) 
		sapi_saprintf(&hkey, "|%s", from_tag);

	return(hkey);
}


static inline u8_t *strans_hash_key(const msg_t *req, bool_t cancel)
{
	u8_t *hkey = NULL, *branch, *sent_by;
	via_t *via = msg_get_topvia(req);

	if(!req->request || !via) {
		STRANS_DEBUG_PRINTF("strans_hash_key(): msg is not a request or first via is missing");
		return(NULL);
	}
	
	branch  = msg_via_getparam(via, "branch");

	if(!branch || !sapi_strncmp(branch, MAGIC_VIA_BRANCH, sizeof(MAGIC_VIA_BRANCH) - 1)) {
		STRANS_DEBUG_PRINTF("strans_hash_key(): msg is not RFC3261 compliant; using compat method");
		hkey = strans_hash_key_compat(req, cancel);
	}
	else {
		sapi_saprintf(&hkey, "%s", branch);
	
		sent_by = msg_via_getparam(via, "sent-by");
		
		if(sent_by) {
			sapi_saprintf(&hkey, "|%s", sent_by);
		}
		
		if(cancel || sapi_strcmp(req->request->method, "ACK"))
			sapi_saprintf(&hkey, "|INVITE");
		else
			sapi_saprintf(&hkey, "|%s", req->request->method);
	}
	
	return(hkey);
}


static void strans_tp_err(const u8_t *reason, u8_t *hkey)
{
        st_t *st;

        if(!__ht || !hkey) {
                STRANS_DEBUG_PRINTF("strans_tp_err(): __ht or hkey is NULL");
                return;
        }

        sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);

        if(!st) {
                STRANS_DEBUG_PRINTF("strans_tp_err(): strans not found");
                return;
        }

        sapi_mutex_lock(&__mutex);
        sapi_hash_del(__ht, hkey);
        sapi_mutex_unlock(&__mutex);
        
        if(st->errcb)
                st->errcb(reason, st->arg);

        st_deref(st);
}


static inline void strans_tx(st_t *st)
{
	if(!st || !st->resp) {
		STRANS_DEBUG_PRINTF("strans_tx(): st or st->resp is NULL");
		return;
	}

	tp_send(st->tp, st->resp, strans_tp_err, st->hkey);
}


static inline void strans_new(const u8_t *hkey, const tp_t *tp, const msg_t *req, strans_req_t reqcb, strans_err_t errcb, void *arg)
{
	via_t *via;
	st_t *st, *ost = NULL;
	ret_t ret;
	tp_t *vtp;

	via = (via_t *)msg_get_topvia(req);

	if(!via) {
		STRANS_DEBUG_PRINTF("strans_new(): via is NULL");
		return;
	}

	vtp = tp_alloc();
	vtp->host   = sapi_strdup(via->host);
	vtp->port   = via->port;
	vtp->proto  = UDP;

	st = st_alloc();

	st->hkey  = sapi_strdup(hkey);
	st->tp    = vtp;
	st->req   = msg_ref(req);
	st->reqcb = reqcb;
	st->errcb = errcb;
	st->arg   = arg;
	
	if(sapi_strcmp(req->request->method, "INVITE")) {
		st->state = PROCEEDING;
		st->resp  = msg_response(req, "100", "Trying");
	}
	else {
		st->state = TRYING;
	}

	sapi_mutex_lock(&__mutex);
	ret = sapi_hash_ins(__ht, hkey, st_ref(st), st_deref);
	sapi_mutex_unlock(&__mutex);

        if(ret != OK) {
                STRANS_DEBUG_PRINTF("strans_new(): sapi_hash_ins(): failed!!!!!");
                st_deref(st);
        }
	else {
		if(sapi_strcmp(req->request->method, "CANCEL")) {
				
			u8_t *ohkey = strans_hash_key(req, TRUE);
			if(ohkey) {
				sapi_mutex_lock(&__mutex);
				ost = st_ref(sapi_hash_find(__ht, ohkey));
				sapi_mutex_unlock(&__mutex);
			}
			sapi_free(ohkey);
		}

		if(st->resp)
			strans_tx(st);
		
		if(st->reqcb) {
			if(ost)
				st->reqcb(st->req, ost->req, st->hkey, ost->hkey, st->arg);
			else
				st->reqcb(st->req, NULL, st->hkey, NULL, st->arg);
		}
	}

	st_deref(st);
	st_deref(ost);
}


static void strans_retrans_i(void *arg, u8_t *hkey)
{
        st_t *st;

        if(!__ht || !hkey) {
                STRANS_DEBUG_PRINTF("strans_retrans_i(): __ht or hkey is NULL");
                return;
        }

        sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);    

        if(!st) {
                // STRANS_DEBUG_PRINTF("strans_retrans_i(): strans not found");
                return;
        }

        if(st_get_state(st) == COMPLETED) {

                strans_tx(st);
		st_inc_ntx(st);
                sapi_oneshot(T1<<st_get_ntx(st), strans_retrans_i, NULL, st->hkey);
        }

        st_deref(st);
}


static void strans_timeout_i(void *arg, u8_t *hkey)
{
        st_t *st;

        if(!__ht || !hkey) {
		STRANS_DEBUG_PRINTF("strans_timeout_i(): __ht or hkey is NULL");
                return;
        }

        sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);    

        if(!st) {
                // STRANS_DEBUG_PRINTF("strans_timeout_i(): strans not found");
                return;
        }
	
        if(st_get_state(st) == COMPLETED) {

		sapi_mutex_lock(&__mutex);
                sapi_hash_del(__ht, hkey);
                sapi_mutex_unlock(&__mutex);

                if(st->errcb)
                        st->errcb("INVITE Request Timeout", st->arg);

        }

        st_deref(st);
}


static void strans_zombie_i(void *arg, u8_t *hkey)
{
        st_t *st;

        if(!__ht || !hkey) {
                STRANS_DEBUG_PRINTF("strans_zombie_i(): __ht or hkey is NULL");
                return;
        }

        sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);    

        if(!st) {
                // STRANS_DEBUG_PRINTF("strans_zombie_i(): strans not found");
                return;
        }
	
        if(st_get_state(st) == CONFIRMED) {

		sapi_mutex_lock(&__mutex);
                sapi_hash_del(__ht, hkey);
                sapi_mutex_unlock(&__mutex);
        }

        st_deref(st);
}


static void strans_zombie_ni(void *arg, u8_t *hkey)
{
        st_t *st;

        if(!__ht || !hkey) {
                STRANS_DEBUG_PRINTF("strans_zombie_i(): __ht or hkey is NULL");
                return;
        }

        sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);    

        if(!st) {
                // STRANS_DEBUG_PRINTF("strans_zombie_i(): strans not found");
                return;
        }
	
        if(st_get_state(st) == COMPLETED) {

		sapi_mutex_lock(&__mutex);
                sapi_hash_del(__ht, hkey);
                sapi_mutex_unlock(&__mutex);
        }

        st_deref(st);
}


static inline void strans_req_i(st_t *st, const msg_t *req)
{
	if(st_get_state(st) == PROCEEDING) {
		
		if(sapi_strcmp(req->request->method, "INVITE")) {
			
			strans_tx(st);
		}
	}
	else if(st_get_state(st) == COMPLETED) {

		if(sapi_strcmp(req->request->method, "INVITE")) {

			strans_tx(st);
		}
		else if(sapi_strcmp(req->request->method, "ACK")) {

			st_set_state(st, CONFIRMED);
			
			if(st->tp->proto == UDP) {
				sapi_oneshot(T4, strans_zombie_i, NULL, st->hkey);
			}
			else {
				sapi_mutex_lock(&__mutex);
				sapi_hash_del(__ht, st->hkey);
				sapi_mutex_unlock(&__mutex);
			}
		}
	}
}


static inline void strans_req_ni(st_t *st)
{
	if(st_get_state(st) == PROCEEDING || st_get_state(st) == COMPLETED) {
		strans_tx(st);
	}
}


static inline void strans_resp_i(st_t *st, const msg_t *resp)
{
	if(st_get_state(st) == PROCEEDING) {

		if(st->resp)
			msg_deref(st->resp);

		st->resp = msg_ref(resp);
		strans_tx(st);

		/* 2XX */
		if(resp->response->statuscode[0] == '2') {

			sapi_mutex_lock(&__mutex);
			sapi_hash_del(__ht, st->hkey);
			sapi_mutex_unlock(&__mutex);			
		}
		/* 300-699 */
		else if(resp->response->statuscode[0] >= '3') {
			
			st_set_state(st, COMPLETED);
			
			if(st->tp->proto == UDP) {
				sapi_oneshot(T1, strans_retrans_i, NULL, st->hkey);
			}

			sapi_oneshot(64 * T1, strans_timeout_i, NULL, st->hkey);
		}
	}
}


static inline void strans_resp_ni(st_t *st, const msg_t *resp)
{
	if(st_get_state(st) == TRYING || st_get_state(st) == PROCEEDING) {

		if(st->resp)
			msg_deref(st->resp);

		st->resp = msg_ref(resp);
		strans_tx(st);

		/* 1XX */
		if(resp->response->statuscode[0] == '1') {

			if(st_get_state(st) == TRYING) {

				st_set_state(st, PROCEEDING);
			}
		}
		/* 200-699 */
		else if(resp->response->statuscode[0] >= '2') {

			st_set_state(st, COMPLETED);

			if(st->tp->proto == UDP) {

				sapi_oneshot(64 * T1, strans_zombie_ni, NULL, st->hkey);
			}
			else {
				sapi_mutex_lock(&__mutex);
				sapi_hash_del(__ht, st->hkey);
				sapi_mutex_unlock(&__mutex);
			}
		}
	}
}


void strans_init(void)
{
	if(__ht)
		return;

	__ht = sapi_hash_alloc(STRANS_HASH_SIZE);
	sapi_logf("Server transaction service is up and running");
}


ret_t strans_req(const tp_t *tp, const msg_t *req, strans_req_t reqcb, strans_err_t errcb, void *arg)
{
	u8_t *hkey = NULL;
	st_t *st;
	ret_t ret = ERR;

	if(!__ht || !tp || !req || !req->request) {
		STRANS_DEBUG_PRINTF("strans_req(): __ht, tp, req or req->request is NULL");
		goto out;
	}

	hkey = strans_hash_key(req, FALSE);
	if(!hkey) {
		STRANS_DEBUG_PRINTF("strans_req(): hkey is NULL");
		goto out;
	}

	sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);

	if(!st) {
		if(sapi_strcmp(req->request->method, "ACK"))
			goto out;
		else
			strans_new(hkey, tp, req, reqcb, errcb, arg);
	}
	else {
		if(sapi_strcmp(st->req->request->method, "INVITE"))
			strans_req_i(st, req);
		else
			strans_req_ni(st);

		st_deref(st);
	}

	ret = OK;

out:
	sapi_free(hkey);
	return(ret);
}


void strans_resp(const u8_t *hkey, const msg_t *resp)
{
        st_t *st;

        if(!__ht || !hkey || !resp || !resp->response) {
                STRANS_DEBUG_PRINTF("strans_resp(): __ht, hkey, resp or resp->response is NULL");
                return;
        }

        sapi_mutex_lock(&__mutex);
        st = st_ref(sapi_hash_find(__ht, hkey));
        sapi_mutex_unlock(&__mutex);

        if(!st) {
                STRANS_DEBUG_PRINTF("strans_resp(): strans not found");
                return;
        }
        
	if(sapi_strcmp(st->req->request->method, "INVITE")) {
		strans_resp_i(st, resp);
	}
	else {
		strans_resp_ni(st, resp);
	}
	
        st_deref(st);
}


void strans_status(void)
{
	if(!__ht)
		return;
	
	sapi_mutex_lock(&__mutex);
	sapi_hash_status(__ht);
	sapi_mutex_unlock(&__mutex);
}


u8_t *strans_status_str(void)
{
	u8_t *str;

	if(!__ht)
		return("strans_status_str(): __ht is NULL");

	sapi_mutex_lock(&__mutex);
	str = sapi_hash_status_str(__ht);
	sapi_mutex_unlock(&__mutex);

	return(str);
}
