/**
 * @file tp.c Transport layer code
 *
 * Copyright 2002-2003 SX Design AS <www.sxdesign.com>. All rights reserved.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>
#include <sapi.h>
#include <msg.h>
#include <tp.h>


typedef struct {
	tp_t *tp;
	u8_t *payload;
	tp_err_t err;
	u8_t *sarg;
} packet_t;


static tp_recv_t _recv   = NULL;
static void *_recv_param = NULL;
static fifo_t *_sendq    = NULL;
static fifo_t *_recvq    = NULL;
static pthread_t _udp_send;
static pthread_t _udp_recv;
static pthread_t _parse;
static u16_t _port = 5060;


static packet_t *packet_alloc()
{
	return((packet_t *)sapi_malloc(sizeof(packet_t)));
}


static void packet_free(packet_t *packet)
{
	if(!packet)
		return;
	
	tp_deref(packet->tp);
	sapi_free(packet->payload);
	sapi_free(packet->sarg);
	sapi_free(packet);
}


static void *parse_thread(void *arg)
{
	while(_recvq) {
		
		packet_t *p = (packet_t *)sapi_fifo_get(_recvq);
		msg_t *msg = msg_alloc();

		// sapi_printf(p->payload);
		
		if(msg_parse(msg, p->payload) == ERR) {
			TP_DEBUG_PRINTF("parse_thread(): Malformed SIP message received");
		}
		else if(_recv) {

			if(msg->request) {
				u8_t *url_str = msg_url_str(msg->request->url);
				
				TP_DEBUG_PRINTF("tp_recv(): Received %s request, uri: '%s'", 
						msg->request->method, url_str);
				
				sapi_free(url_str);
			}
			else if(msg->response) {
				TP_DEBUG_PRINTF("tp_recv(): Received %s (%s) response", 
						msg->response->statuscode, msg->response->reasonphrase);
			}

			_recv(p->tp, msg, _recv_param);
		}
		msg_deref(msg);
		packet_free(p);
	}

	TP_DEBUG_PRINTF("parse_thread(): exit");

	return(NULL);
}


static void *udp_recv_thread(void *arg)
{
	int size, on = 1;
	struct sockaddr_in sa, ca;
	socklen_t ca_len = sizeof(ca);
	char buf[64000];
	int fd;

	bzero(&(sa), sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(_port);

	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		sapi_logf("udp_recv_thread(): socket(): %s", strerror(errno));
		goto out;
	}

	if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
		sapi_logf("udp_listen_thread(): setsockopt(): %s", strerror(errno));
		goto out;
	}

	if(bind(fd, (struct sockaddr *)&sa, (socklen_t)sizeof(sa)) < 0) {
		sapi_logf("udp_listen_thread(): bind(): Port %u: Errno: %i: %s", _port, errno, strerror(errno));
		goto out;
	}
	
	while(_recvq) {
		packet_t *packet;
#if 1
		size = recvfrom(fd, buf, sizeof(buf) - 1, 0, (struct sockaddr *)&ca, &ca_len);
		buf[size] = '\0';

		packet = packet_alloc();
		packet->payload = sapi_strdup(buf);
		packet->tp = tp_alloc();

		packet->tp->host  = sapi_strdup("todo.com");
		packet->tp->port  = 5060;
		packet->tp->proto = UDP;

		sapi_fifo_put(_recvq, (void *)packet);
#endif
		// TP_DEBUG_PRINTF("udp_recv_thread(): Packet received");
	}

 out:
	TP_DEBUG_PRINTF("udp_recv_thread(): exit");

	if(fd != -1)
		close(fd);

	return(NULL);
}


static void *udp_send_thread(void *arg)
{
	int fd, on = 1;
	struct sockaddr_in sa;

	bzero(&(sa), sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(_port);
	
	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		sapi_logf("udp_send_thread(): socket(): %s", strerror(errno));
		goto out;
	}

	if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
		sapi_logf("udp_send_thread(): setsockopt(): %s", strerror(errno));
		goto out;
	}
	
	if(bind(fd, (struct sockaddr *)&sa, (socklen_t)sizeof(sa)) < 0) {
		sapi_logf("udp_send_thread(): bind(): Port %u: Errno: %i: %s", _port, errno, strerror(errno));
		goto out;
	}

	while(_sendq) {

		packet_t *p = (packet_t *)sapi_fifo_get(_sendq);
		struct hostent hent, *hres;
		u8_t hbuf[8192];
		s32_t h_errno;
		u32_t addr;
		struct sockaddr_in sa;
		
		int ret = gethostbyname_r(p->tp->host, &hent, hbuf, sizeof(hbuf), &hres, &h_errno);
		if(ret || (hres == NULL)) {
			if(p->err) {
				u8_t *errmsg = sapi_sprintf("%s: %s", hstrerror(h_errno), p->tp->host);
				p->err(errmsg, p->sarg);
				sapi_free(errmsg);
			}
			goto err;
		}

		bcopy(hres->h_addr_list[0], (char *)&addr, sizeof(addr));
		endhostent();

		bzero((void *)&sa, sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_port   = htons((unsigned short int)p->tp->port);
		sa.sin_addr.s_addr = addr;

		if(sendto(fd, (void *)p->payload, (size_t)sapi_strlen(p->payload), 
			  0, (struct sockaddr *)&sa, (socklen_t)sizeof(sa)) < 0) {

			if(p->err) {
				p->err(strerror(errno), p->sarg);
			}
			goto err;
		}
		
		// TP_DEBUG_PRINTF("udp_send_thread(): packet sent to %s", p->tp->host);
		
err:
		packet_free(p);
	}

out:	
	if(fd != -1)
		close(fd);

	TP_DEBUG_PRINTF("udp_send_thread(): exit");
	
	return(NULL);
}


static void tp_free(tp_t *tp)
{
	if(!tp)
		return;

	sapi_free(tp->host);
	sapi_free(tp);
}


void tp_init(u16_t port, tp_recv_t recv, void *param)
{
	_recv       = recv;
	_recv_param = param;
	_sendq      = sapi_fifo_alloc();
	_recvq      = sapi_fifo_alloc();
	_port       = port;

	pthread_create(&_udp_send, NULL, udp_send_thread, NULL);
	pthread_create(&_udp_recv, NULL, udp_recv_thread, NULL);
	pthread_create(&_parse, NULL, parse_thread, NULL);

	TP_DEBUG_PRINTF("Transport service is up and running");
}


tp_t *tp_alloc(void)
{
	tp_t *tp = (tp_t *)sapi_malloc(sizeof(tp_t));

	tp->nrefs = 1;
	sapi_mutex_init(&tp->mutex);

	return(tp);
}


const tp_t *tp_ref(const tp_t *tp)
{
	if(!tp)
		return(NULL);

	sapi_mutex_lock(&((tp_t *)tp)->mutex);
	((tp_t *)tp)->nrefs++;
	sapi_mutex_unlock(&((tp_t *)tp)->mutex);

	return(tp);
}


void tp_deref(const tp_t *tp)
{
	u32_t nrefs;

	if(!tp)
		return;

	sapi_mutex_lock(&((tp_t *)tp)->mutex);
	nrefs = --((tp_t *)tp)->nrefs;
	sapi_mutex_unlock(&((tp_t *)tp)->mutex);

	if(!nrefs)
		tp_free((tp_t *)tp);
}


void tp_send(const tp_t *tp, const msg_t *msg, tp_err_t err, const u8_t *sarg)
{
	u8_t *payload;
	packet_t *packet;

	if(!_sendq) {
		TP_DEBUG_PRINTF("tp_send(): send queue not initialized");
		return;
	}
	
	payload = msg_str(msg);
	if(!payload) {
		TP_DEBUG_PRINTF("tp_send(): msg_str(): failed");
		return;
	}

	packet = packet_alloc();

	packet->tp      = (tp_t *)tp_ref(tp);
	packet->payload = payload;
	packet->err     = err;
	packet->sarg    = sapi_strdup(sarg);

	if(msg->request) {
		u8_t *url_str = msg_url_str(msg->request->url);

		TP_DEBUG_PRINTF("tp_send(): Sending %s request, uri: '%s' [%s:%i]", 
				msg->request->method, url_str, tp->host, tp->port);

		sapi_free(url_str);
	}
	else if(msg->response) {
		TP_DEBUG_PRINTF("tp_send(): Sending %s (%s) response [%s:%i]", 
				msg->response->statuscode, msg->response->reasonphrase,
				tp->host, tp->port);
	}

	sapi_fifo_put(_sendq, (void *)packet);
}


void tp_status(void)
{
	if(!_sendq || !_recvq)
		return;

	sapi_printf(" Send Queue");
	sapi_fifo_status(_sendq);
	sapi_printf(" Recv Queue");
	sapi_fifo_status(_recvq);
}


u8_t *tp_status_str(void)
{
	u8_t *str = NULL, *tmp;

	if(!_sendq || !_recvq)
		return(sapi_sprintf("tp_status_str(): _sendq or recvq is NULL"));

	sapi_saprintf(&str, " Send Queue\n");
	tmp = sapi_fifo_status_str(_sendq);
	sapi_saprintf(&str, "%s", tmp);
	sapi_free(tmp);

	sapi_saprintf(&str, " Recv Queue\n");
	tmp = sapi_fifo_status_str(_recvq);
	sapi_saprintf(&str, "%s", tmp);
	sapi_free(tmp);

	return(str);
}


u8_t *tp_status_str_recv(void)
{
	if(!_recvq)
		return(sapi_sprintf("tp_status_str_recv(): _recvq is NULL"));

	return(sapi_fifo_status_str(_recvq));
}


u8_t *tp_status_str_send(void)
{
	if(!_sendq)
		return(sapi_sprintf("tp_status_str_send(): _sendq is NULL"));

	return(sapi_fifo_status_str(_sendq));
}
